import asyncio
import logging
import struct
import typing
from types import MappingProxyType

import msgspec

from aiomisc.entrypoint import entrypoint
from aiomisc.service import UDPServer

from .spec import Error, Payload, Request, Response

log = logging.getLogger()


class RPCCallError(RuntimeError):
    pass


class RPCServer(UDPServer):
    __required__ = ("handlers",)

    HEADER = struct.Struct(">I")
    handlers: dict[str, typing.Callable]

    _serial: int
    _encoder: msgspec.msgpack.Encoder
    _decoder: msgspec.msgpack.Decoder
    _futures: typing.MutableMapping[int, asyncio.Future]

    async def start(self) -> None:
        self._serial: int = 0
        self._futures = {}
        self._decoder = msgspec.msgpack.Decoder(Payload)
        self._encoder = msgspec.msgpack.Encoder()
        await super().start()

    async def execute(self, method: str, kwargs: dict):
        func = self.handlers[method]

        if asyncio.iscoroutinefunction(func):
            return await func(**kwargs)
        else:
            return func(**kwargs)

    async def _on_request(self, request: Request, addr: tuple) -> typing.Any:
        try:
            response = Response(
                id=request.id,
                result=await self.execute(request.method, request.params),
            )
        except Exception as e:
            response = Response(
                id=request.id, error=Error(type=str(type(e)), args=e.args)
            )

        self.sendto(self._encoder.encode(Payload(response=response)), addr)

    def _on_response(self, response: Response) -> typing.Any:
        future: asyncio.Future = self._futures.pop(response.id)
        if not future.done():
            if response.result is not None:
                future.set_result(response.result)
                return

            future.set_exception(
                RPCCallError(response.error.type, response.error.args)
            )

    async def handle_datagram(self, data: bytes, addr):
        body_bytes = data
        payload: Payload = self._decoder.decode(body_bytes)

        if payload.request is not None:
            self.create_task(self._on_request(payload.request, addr))
        elif payload.response is not None:
            self._on_response(payload.response)

    def __call__(
        self, host: str, port: int, method: str, **params: typing.Any
    ) -> typing.Any:
        self._serial += 1
        serial = self._serial
        future = self.loop.create_future()
        request: bytes = self._encoder.encode(
            Payload(request=Request(id=serial, method=method, params=params))
        )
        self._futures[serial] = future
        self.sendto(request, (host, port))
        return future


handlers = MappingProxyType({"multiply": lambda x, y: x * y})


if __name__ == "__main__":
    service = RPCServer(handlers=handlers, address="::", port=15678)

    with entrypoint(service) as loop:
        loop.run_forever()



import socket

import dnslib  # type: ignore[import-untyped]
import pytest

from aiomisc import threaded
from aiomisc.service.dns import DNSStore, DNSZone, TCPDNSServer, UDPDNSServer
from aiomisc.service.dns.records import AAAA, CNAME, A, RecordType
from aiomisc.service.dns.service import TCP_HEADER_STRUCT


@pytest.fixture
def dns_store_filled():
    store = DNSStore()
    zone = DNSZone("example.com.")
    a_record = A.create(name="sub.example.com.", ip="192.0.2.1", ttl=3600)
    aaaa_record = AAAA.create(
        name="ipv6.example.com.", ipv6="2001:db8::1", ttl=3600
    )
    cname_record = CNAME.create(name="alias.example.com.", label="example.com.")
    zone.add_record(a_record)
    zone.add_record(aaaa_record)
    zone.add_record(cname_record)
    store.add_zone(zone)
    return store


@pytest.fixture
def dns_server_port(aiomisc_unused_port_factory) -> int:
    return aiomisc_unused_port_factory()


@pytest.fixture
def dns_server_udp(dns_store_filled: DNSStore, dns_server_port) -> UDPDNSServer:
    server = UDPDNSServer(
        store=dns_store_filled, address="localhost", port=dns_server_port
    )
    return server


@pytest.fixture
def dns_server_tcp(dns_store_filled: DNSStore, dns_server_port) -> TCPDNSServer:
    server = TCPDNSServer(
        store=dns_store_filled, address="localhost", port=dns_server_port
    )
    return server


@pytest.fixture
def services(dns_server_udp, dns_server_tcp):
    return [dns_server_udp, dns_server_tcp]


@threaded
def dns_send_receive_udp(data, port):
    # Send the query
    with socket.socket(socket.AF_INET, socket.SOCK_DGRAM) as sock:
        sock.settimeout(5)
        sock.sendto(data, ("localhost", port))
        # Receive the response
        response_data, _ = sock.recvfrom(65535)
        return dnslib.DNSRecord.parse(response_data)


@threaded
def dns_send_receive_tcp(data, port):
    # Send the query
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock:
        sock.settimeout(5)
        sock.connect(("localhost", port))
        sock.sendall(data)
        # Receive the response length (2 bytes)
        length_data = sock.recv(2)
        length = int.from_bytes(length_data, byteorder="big")
        # Receive the actual response
        response_data = sock.recv(length)
        return dnslib.DNSRecord.parse(response_data)


async def test_handle_datagram_a_record(services, dns_server_port):
    # Prepare a DNS query for A record
    query = dnslib.DNSRecord.question("sub.example.com.", qtype="A")
    query_data = query.pack()

    response = await dns_send_receive_udp(query_data, dns_server_port)

    # Verify the response
    assert response.header.rcode == dnslib.RCODE.NOERROR
    assert len(response.rr) == 1
    assert str(response.rr[0].rname) == "sub.example.com."
    assert response.rr[0].rdata == dnslib.A("192.0.2.1")


async def test_handle_datagram_aaaa_record(services, dns_server_port):
    # Prepare a DNS query for AAAA record
    query = dnslib.DNSRecord.question("ipv6.example.com.", qtype="AAAA")
    query_data = query.pack()

    response = await dns_send_receive_udp(query_data, dns_server_port)

    # Verify the response
    assert response.header.rcode == dnslib.RCODE.NOERROR
    assert len(response.rr) == 1
    assert str(response.rr[0].rname) == "ipv6.example.com."
    assert response.rr[0].rdata == dnslib.AAAA("2001:db8::1")


async def test_handle_datagram_cname_record(services, dns_server_port):
    # Prepare a DNS query for CNAME record
    query = dnslib.DNSRecord.question("alias.example.com.", qtype="CNAME")
    query_data = query.pack()

    response = await dns_send_receive_udp(query_data, dns_server_port)

    # Verify the response
    assert response.header.rcode == dnslib.RCODE.NOERROR
    assert len(response.rr) == 1
    assert str(response.rr[0].rname) == "alias.example.com."
    assert response.rr[0].rdata == dnslib.CNAME("example.com.")


async def test_handle_datagram_nonexistent_record(services, dns_server_port):
    # Prepare a DNS query for a nonexistent record
    query = dnslib.DNSRecord.question("nonexistent.example.com.", qtype="A")
    query_data = query.pack()

    response = await dns_send_receive_udp(query_data, dns_server_port)

    # Verify the response
    assert response.header.rcode == dnslib.RCODE.NXDOMAIN
    assert len(response.rr) == 0


async def test_handle_datagram_remove_record(
    services, dns_store_filled, dns_server_port
):
    # Remove an existing record from the zone
    zone = dns_store_filled.get_zone("example.com.")

    for record in zone.get_records("sub.example.com.", RecordType.A):
        zone.remove_record(record)

    # Prepare a DNS query for the removed record
    query = dnslib.DNSRecord.question("sub.example.com.", qtype="A")
    query_data = query.pack()

    response = await dns_send_receive_udp(query_data, dns_server_port)

    # Verify the response
    assert response.header.rcode == dnslib.RCODE.NXDOMAIN
    assert len(response.rr) == 0


async def test_handle_datagram_edns_record(services, dns_server_port):
    # Prepare a DNS query with EDNS0
    query = dnslib.DNSRecord.question("sub.example.com.", qtype="A")
    query.add_ar(dnslib.EDNS0(udp_len=4096))
    query_data = query.pack()

    response = await dns_send_receive_udp(query_data, dns_server_port)

    # Verify the response
    assert response.header.rcode == dnslib.RCODE.NOERROR
    assert len(response.rr) == 1
    assert str(response.rr[0].rname) == "sub.example.com."
    assert response.rr[0].rdata == dnslib.A("192.0.2.1")
    assert response.ar[0].rtype == dnslib.QTYPE.OPT


async def test_handle_large_datagram_truncated_udp(
    services, dns_server_port, dns_store_filled
):
    # Add many A records to the store to exceed typical UDP packet size
    zone = dns_store_filled.get_zone("example.com.")
    for i in range(1, 101):
        a_record = A.create(name="rr.example.com.", ip=f"192.0.2.{i}")
        zone.add_record(a_record)

    # Prepare a DNS query
    query = dnslib.DNSRecord.question("rr.example.com.", qtype="A")
    query_data = query.pack()

    response = await dns_send_receive_udp(query_data, dns_server_port)

    # Verify the response is truncated
    assert response.header.tc == 1  # Ensure the TC (truncated) bit is set


async def test_handle_large_tcp_request(
    services, dns_server_port, dns_store_filled
):
    # Add many A records to the store to exceed typical UDP packet size
    zone = dns_store_filled.get_zone("example.com.")
    for i in range(1, 101):
        a_record = A.create(name="rr.example.com.", ip=f"192.0.2.{i}", ttl=3600)
        zone.add_record(a_record)

    # Prepare a DNS query for TCP
    query = dnslib.DNSRecord.question("rr.example.com.", qtype="A")
    query_data = TCP_HEADER_STRUCT.pack(len(query.pack())) + query.pack()

    response = await dns_send_receive_tcp(query_data, dns_server_port)

    # Verify the TCP response is not truncated and contains all records
    assert response.header.rcode == dnslib.RCODE.NOERROR
    assert len(response.rr) == 100
    assert all(str(rr.rname).startswith("rr") for rr in response.rr)



import asyncio
from types import MappingProxyType

from aiomisc.entrypoint import entrypoint
from aiomisc.service import TCPServer

from .spec import RPCBase


class RPCServer(TCPServer, RPCBase):
    async def handle_client(
        self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter
    ):
        await self.communicate(reader, writer)

    async def start(self) -> None:
        await RPCBase.start(self)
        await TCPServer.start(self)


handlers = MappingProxyType({"multiply": lambda x, y: x * y})


if __name__ == "__main__":
    service = RPCServer(handlers=handlers, address="::", port=5678)

    with entrypoint(service) as loop:
        loop.run_forever()



import asyncio
import logging
import re
import sys
from collections import defaultdict
from collections.abc import Mapping, Sequence
from concurrent.futures import Executor
from types import MappingProxyType
from typing import Any

from .base import Service

try:
    import grpc.aio
    from grpc_reflection.v1alpha import reflection
except ImportError as e:
    raise ImportError(
        "You must install 'grpcio' manually or using extras 'aiomisc[grpc]'"
    ) from e

log = logging.getLogger(__name__)

PortFuture = asyncio.Future[int]


class GRPCService(Service):
    GRACEFUL_STOP_TIME: float = 60.0

    _ADDRESS_REGEXP = re.compile(
        r"(?P<address>(\[((([([0-9a-fA-F:]*)+)])?|([\w.]+))):(\d+)"
    )

    _server: grpc.aio.Server
    _server_args: MappingProxyType
    _insecure_ports: set[tuple[str, PortFuture]]
    _secure_ports: set[tuple[str, grpc.ServerCredentials, PortFuture]]
    _registered_services: defaultdict[str, dict[str, grpc.RpcMethodHandler]]

    def __init__(
        self,
        *,
        migration_thread_pool: Executor | None = None,
        handlers: Sequence[grpc.ServiceRpcHandler] | None = None,
        interceptors: Sequence[Any] | None = None,
        options: Sequence[tuple[str, Any]] | None = None,
        maximum_concurrent_rpcs: int | None = None,
        compression: grpc.Compression | None = None,
        reflection: bool = False,
        **kwds: Any,
    ):
        self._server_args = MappingProxyType(
            {
                "compression": compression,
                "handlers": handlers,
                "interceptors": interceptors,
                "maximum_concurrent_rpcs": maximum_concurrent_rpcs,
                "migration_thread_pool": migration_thread_pool,
                "options": options,
            }
        )
        self._services: set[grpc.ServiceRpcHandler] = set()
        self._insecure_ports = set()
        self._secure_ports = set()
        self._reflection = reflection
        self._registered_services = defaultdict(dict)
        super().__init__(**kwds)

    @classmethod
    def _log_port(cls, msg: str, address: str, bind_port: Any) -> None:
        match: re.Match | None = cls._ADDRESS_REGEXP.match(address)

        if match is not None:
            groups = match.groupdict()
            address = groups["address"]

        log.info("%s: grpc://%s:%s", msg, address, bind_port)

    async def start(self) -> None:
        self._server = grpc.aio.server(**self._server_args)

        for address, future in self._insecure_ports:
            port = self._server.add_insecure_port(address)
            future.set_result(port)
            self._log_port("Listening insecure address", address, port)

        for address, credentials, future in self._secure_ports:
            port = self._server.add_secure_port(address, credentials)
            future.set_result(port)
            self._log_port("Listening secure address", address, port)

        if self._reflection:
            service_names = [x.service_name() for x in self._services]
            service_names.append(reflection.SERVICE_NAME)
            reflection.enable_server_reflection(service_names, self._server)

        for name, handlers in self._registered_services.items():
            # noinspection PyUnresolvedReferences
            self._server.add_registered_method_handlers(  # type: ignore
                name, handlers
            )

        self._server.add_generic_rpc_handlers(tuple(self._services))
        await self._server.start()

    async def stop(self, exception: Exception | None = None) -> None:
        await self._server.stop(self.GRACEFUL_STOP_TIME)

    def add_generic_rpc_handlers(
        self, generic_rpc_handlers: Sequence[grpc.ServiceRpcHandler]
    ) -> None:
        for service in generic_rpc_handlers:
            self._services.add(service)

    def add_registered_method_handlers(
        self, name: str, handlers: Mapping[str, grpc.RpcMethodHandler]
    ) -> None:
        self._registered_services[name].update(handlers)

    def add_insecure_port(self, address: str) -> PortFuture:
        future: PortFuture = asyncio.Future()
        self._insecure_ports.add((address, future))
        return future

    def add_secure_port(
        self, address: str, server_credentials: grpc.ServerCredentials
    ) -> PortFuture:
        future: PortFuture = asyncio.Future()
        self._secure_ports.add((address, server_credentials, future))
        return future



import asyncio

import pytest

import aiomisc

pytestmark = pytest.mark.catch_loop_exceptions


async def test_periodic(event_loop):
    condition = asyncio.Condition()
    counter = 0

    async def task():
        nonlocal counter
        counter += 1

        async with condition:
            condition.notify_all()

    periodic = aiomisc.PeriodicCallback(task)
    periodic.start(0.1, event_loop)

    async with condition:
        await asyncio.wait_for(
            condition.wait_for(lambda: counter >= 5), timeout=5
        )

    with pytest.raises(asyncio.CancelledError):
        await periodic.stop()


async def test_periodic_return_exceptions(event_loop):
    condition = asyncio.Condition()
    counter = 0

    async def task():
        nonlocal counter
        counter += 1

        async with condition:
            condition.notify_all()

    periodic = aiomisc.PeriodicCallback(task)
    periodic.start(0.1, event_loop)

    async with condition:
        await asyncio.wait_for(
            condition.wait_for(lambda: counter >= 5), timeout=5
        )

    await periodic.stop(return_exceptions=True)


async def test_long_func(event_loop):
    counter = 0
    condition = asyncio.Condition()

    async def task():
        nonlocal counter
        async with condition:
            counter += 1
            await asyncio.sleep(0.5)
            condition.notify_all()

    periodic = aiomisc.PeriodicCallback(task)
    periodic.start(0.1, event_loop)

    await asyncio.sleep(1)
    with pytest.raises(asyncio.CancelledError):
        await periodic.stop()

    async with condition:
        await asyncio.wait_for(
            condition.wait_for(lambda: counter == 2), timeout=2
        )


async def test_shield(event_loop):
    counter = 0

    async def task():
        nonlocal counter
        await asyncio.sleep(0.1)
        counter += 1

    periodic = aiomisc.PeriodicCallback(task)
    periodic.start(0.2, event_loop, shield=True)

    # Wait for periodic callback to start
    await asyncio.sleep(0.01)

    with pytest.raises(asyncio.CancelledError):
        await periodic.stop()

    # Wait for counter to increment
    await asyncio.sleep(0.1)

    # Shielded
    assert counter == 1

    # No shield
    counter = 0
    periodic = aiomisc.PeriodicCallback(task)
    periodic.start(0.2, event_loop, shield=False)

    # Wait for periodic callback to start
    await asyncio.sleep(0.01)

    with pytest.raises(asyncio.CancelledError):
        await periodic.stop()

    # Wait for counter to increment
    await asyncio.sleep(0.1)

    # Cancelled
    assert counter == 0


async def test_delay(event_loop):
    counter = 0

    def task():
        nonlocal counter
        counter += 1

    periodic = aiomisc.PeriodicCallback(task)
    periodic.start(0.1, event_loop, delay=0.5)

    await asyncio.sleep(0.25)

    assert not counter

    await asyncio.sleep(0.5)

    with pytest.raises(asyncio.CancelledError):
        await periodic.stop()

    assert 1 < counter <= 4


async def test_restart(event_loop):
    counter = 0
    condition = asyncio.Condition()

    async def task():
        nonlocal counter
        async with condition:
            counter += 1
            condition.notify_all()

    periodic = aiomisc.PeriodicCallback(task)

    for i in (5, 10, 15):
        periodic.start(0.1, event_loop)

        async with condition:
            await asyncio.wait_for(
                condition.wait_for(lambda: counter == i), timeout=5
            )

        with pytest.raises(asyncio.CancelledError):
            await periodic.stop()

        assert counter == i


@aiomisc.timeout(10)
async def test_cancelled_callback(event_loop):
    counter = 0
    condition = asyncio.Condition()

    async def task():
        nonlocal counter
        async with condition:
            counter += 1
            condition.notify_all()
        raise asyncio.CancelledError

    periodic = aiomisc.PeriodicCallback(task)

    for i in (5, 10, 15):
        periodic.start(0.1, event_loop)

        async with condition:
            await asyncio.wait_for(
                condition.wait_for(lambda: counter == i), timeout=5
            )

        with pytest.raises(asyncio.CancelledError):
            await periodic.stop()

        assert counter == i



from typing import Any, Dict, Optional

import msgspec


class Request(msgspec.Struct):
    id: int
    method: str
    params: dict[str, Any]


class Error(msgspec.Struct):
    type: str
    args: Any


class Response(msgspec.Struct):
    id: int
    result: Any | None = None
    error: Error | None = None


class Payload(msgspec.Struct):
    request: Request | None = None
    response: Response | None = None



import asyncio
import functools
import inspect
import logging
from asyncio import CancelledError, Event, Future, Lock, wait_for
from collections.abc import Callable, Coroutine, Iterable
from dataclasses import dataclass
from inspect import Parameter
from typing import Any, Generic, Protocol, TypeVar

from .compat import EventLoopMixin
from .counters import Statistic

log = logging.getLogger(__name__)


V = TypeVar("V")
R = TypeVar("R")


@dataclass(frozen=True)
class Arg(Generic[V, R]):
    value: V
    future: "Future[R]"


class ResultNotSetError(Exception):
    pass


class AggregateAsyncFunc(Protocol, Generic[V, R]):
    __name__: str

    async def __call__(self, *args: Arg[V, R]) -> None: ...


class AggregateStatistic(Statistic):
    leeway_ms: float
    max_count: int
    success: int
    error: int
    done: int


def _has_variadic_positional(func: Callable[..., Any]) -> bool:
    return any(
        parameter.kind == Parameter.VAR_POSITIONAL
        for parameter in inspect.signature(func).parameters.values()
    )


class AggregatorAsync(EventLoopMixin, Generic[V, R]):
    _func: AggregateAsyncFunc[V, R]
    _max_count: int | None
    _leeway: float
    _first_call_at: float | None
    _args: list
    _futures: "list[Future[R]]"
    _event: Event
    _lock: Lock

    def __init__(
        self,
        func: AggregateAsyncFunc[V, R],
        *,
        leeway_ms: float,
        max_count: int | None = None,
        statistic_name: str | None = None,
    ):
        if not _has_variadic_positional(func):
            raise ValueError(
                "Function must accept variadic positional arguments"
            )

        if max_count is not None and max_count <= 0:
            raise ValueError("max_count must be positive int or None")

        if leeway_ms <= 0:
            raise ValueError("leeway_ms must be positive float")

        self._func = func
        self._max_count = max_count
        self._leeway = leeway_ms / 1000
        self._clear()
        self._statistic = AggregateStatistic(statistic_name)
        self._statistic.leeway_ms = self.leeway_ms
        self._statistic.max_count = max_count or 0

    def _clear(self) -> None:
        self._first_call_at = None
        self._args = []
        self._futures = []
        self._event = Event()
        self._lock = Lock()

    @property
    def max_count(self) -> int | None:
        return self._max_count

    @property
    def leeway_ms(self) -> float:
        return self._leeway * 1000

    @property
    def count(self) -> int:
        return len(self._args)

    async def _execute(
        self, *, args: list[V], futures: "list[Future[R]]"
    ) -> None:
        args_ = [
            Arg(value=arg, future=future) for arg, future in zip(args, futures)
        ]
        try:
            await self._func(*args_)
            self._statistic.success += 1
        except CancelledError:
            # Other waiting tasks can try to finish the job instead.
            raise
        except Exception as e:
            self._set_exception(e, futures)
            self._statistic.error += 1
            return
        finally:
            self._statistic.done += 1

        # Validate that all results/exceptions are set by the func
        for future in futures:
            if not future.done():
                future.set_exception(ResultNotSetError)

    def _set_exception(
        self, exc: Exception, futures: list["Future[R]"]
    ) -> None:
        for future in futures:
            if not future.done():
                future.set_exception(exc)

    async def aggregate(self, arg: V) -> R:
        if self._first_call_at is None:
            self._first_call_at = self.loop.time()
        first_call_at = self._first_call_at

        args: list = self._args
        futures: list[Future[R]] = self._futures
        event: Event = self._event
        lock: Lock = self._lock
        args.append(arg)
        future: Future[R] = Future()
        futures.append(future)

        if self.count == self.max_count:
            event.set()
            self._clear()
        else:
            # Waiting for max_count requests or a timeout
            try:
                await wait_for(
                    event.wait(),
                    timeout=first_call_at + self._leeway - self.loop.time(),
                )
            except asyncio.TimeoutError:
                log.debug(
                    "Aggregation timeout of %s for batch started at %.4f "
                    "with %d calls after %.2f ms",
                    self._func.__name__,
                    first_call_at,
                    len(futures),
                    (self.loop.time() - first_call_at) * 1000,
                )

        # Clear only if not cleared already
        if args is self._args:
            self._clear()

        # Trying to acquire the lock to execute the aggregated function
        async with lock:
            if not future.done():
                await self._execute(args=args, futures=futures)
        await future
        return future.result()


S = TypeVar("S", contravariant=True)
T = TypeVar("T", covariant=True)


class AggregateFunc(Protocol, Generic[S, T]):
    __name__: str

    async def __call__(self, *args: S) -> Iterable[T]: ...


def _to_async_aggregate(func: AggregateFunc[V, R]) -> AggregateAsyncFunc[V, R]:
    @functools.wraps(
        func,
        assigned=tuple(
            item
            for item in functools.WRAPPER_ASSIGNMENTS
            if item != "__annotations__"
        ),
    )
    async def wrapper(*args: Arg[V, R]) -> None:
        args_ = [item.value for item in args]
        results = await func(*args_)
        for res, arg in zip(results, args):
            if not arg.future.done():
                arg.future.set_result(res)

    return wrapper


class Aggregator(AggregatorAsync[V, R], Generic[V, R]):
    def __init__(
        self,
        func: AggregateFunc[V, R],
        *,
        leeway_ms: float,
        max_count: int | None = None,
        statistic_name: str | None = None,
    ) -> None:
        if not _has_variadic_positional(func):
            raise ValueError(
                "Function must accept variadic positional arguments"
            )

        super().__init__(
            _to_async_aggregate(func),
            leeway_ms=leeway_ms,
            max_count=max_count,
            statistic_name=statistic_name,
        )


def aggregate(
    leeway_ms: float, max_count: int | None = None
) -> Callable[[AggregateFunc[V, R]], Callable[[V], Coroutine[Any, Any, R]]]:
    """
    Parametric decorator that aggregates multiple
    (but no more than ``max_count`` defaulting to ``None``) single-argument
    executions (``res1 = await func(arg1)``, ``res2 = await func(arg2)``, ...)
    of an asynchronous function with variadic positional arguments
    (``async def func(*args, pho=1, bo=2) -> Iterable``) into its single
    execution with multiple positional arguments
    (``res1, res2, ... = await func(arg1, arg2, ...)``) collected within a time
    window ``leeway_ms``.

    .. note::

        ``func`` must return a sequence of values of length equal to the
        number of arguments (and in the same order).

    .. note::

        if some unexpected error occurs, exception is propagated to each
        future; to set an individual error for each aggregated call refer
        to ``aggregate_async``.

    :param leeway_ms: The maximum approximate delay between the first
           collected argument and the aggregated execution.
    :param max_count: The maximum number of arguments to call decorated
           function with. Default ``None``.

    :return:
    """

    def decorator(
        func: AggregateFunc[V, R],
    ) -> Callable[[V], Coroutine[Any, Any, R]]:
        aggregator = Aggregator(func, max_count=max_count, leeway_ms=leeway_ms)
        return aggregator.aggregate

    return decorator


def aggregate_async(
    leeway_ms: float, max_count: int | None = None
) -> Callable[
    [AggregateAsyncFunc[V, R]], Callable[[V], Coroutine[Any, Any, R]]
]:
    """
    Same as ``aggregate``, but with ``func`` arguments of type ``Arg``
    containing ``value`` and ``future`` attributes instead. In this setting
    ``func`` is responsible for setting individual results/exceptions for all
    of the futures or throwing an exception (it will propagate to futures
    automatically). If ``func`` mistakenly does not set a result of some
    future, then, ``ResultNotSetError`` exception is set.

    :return:
    """

    def decorator(
        func: AggregateAsyncFunc[V, R],
    ) -> Callable[[V], Coroutine[Any, Any, R]]:
        aggregator = AggregatorAsync(
            func, max_count=max_count, leeway_ms=leeway_ms
        )
        return aggregator.aggregate

    return decorator



import asyncio
from unittest.mock import patch

import pytest

from aiomisc.cron import CronCallback

# Fixed delay for deterministic testing
TICK_DELAY = 0.1


def mock_get_next(*args):
    """Return a fixed delay for deterministic testing."""
    return TICK_DELAY


@patch.object(CronCallback, "get_next", mock_get_next)
async def test_cron(event_loop):
    counter = 0

    def task():
        nonlocal counter
        counter += 1

    cron = CronCallback(task)
    cron.start("* * * * * *", event_loop)

    # Wait for at least 2 ticks
    await asyncio.sleep(TICK_DELAY * 2.5)
    with pytest.raises(asyncio.CancelledError):
        await cron.stop()

    assert counter == 2

    await asyncio.sleep(TICK_DELAY * 2)

    # Counter should not increase after stop
    assert counter == 2


@patch.object(CronCallback, "get_next", mock_get_next)
async def test_long_func(event_loop):
    counter = 0
    condition = asyncio.Condition()

    async def task():
        nonlocal counter
        async with condition:
            await asyncio.sleep(TICK_DELAY * 2)  # Longer than tick interval
            counter += 1
            condition.notify_all()

    cron = CronCallback(task)
    cron.start("* * * * * *", event_loop)

    async with condition:
        await condition.wait_for(lambda: counter >= 1)

    with pytest.raises(asyncio.CancelledError):
        await cron.stop()

    assert counter == 1


@patch.object(CronCallback, "get_next", mock_get_next)
async def test_shield(event_loop):
    counter = 0
    condition = asyncio.Condition()

    async def task():
        nonlocal counter, condition
        counter += 1
        async with condition:
            condition.notify_all()
        await asyncio.sleep(TICK_DELAY * 2)
        counter += 1

    cron = CronCallback(task)
    cron.start("* * * * * *", event_loop, shield=True)

    # Wait for cron callback to start
    async with condition:
        await condition.wait_for(lambda: counter >= 1)

    with pytest.raises(asyncio.CancelledError):
        await cron.stop()

    await asyncio.sleep(TICK_DELAY * 3)

    # Shielded - task should complete
    assert counter == 2


@patch.object(CronCallback, "get_next", mock_get_next)
async def test_restart(event_loop):
    counter = 0
    condition = asyncio.Condition()

    async def task():
        nonlocal counter, condition
        counter += 1
        async with condition:
            condition.notify_all()

    cron = CronCallback(task)
    cron.start("* * * * * *", event_loop)

    async with condition:
        await condition.wait_for(lambda: counter >= 2)

    with pytest.raises(asyncio.CancelledError):
        await cron.stop()

    assert counter == 2

    await asyncio.sleep(TICK_DELAY * 2)

    # Counter should not increase after stop
    assert counter == 2

    cron.start("* * * * * *", event_loop)

    async with condition:
        await condition.wait_for(lambda: counter >= 4)

    with pytest.raises(asyncio.CancelledError):
        await cron.stop()

    assert counter == 4

    await asyncio.sleep(TICK_DELAY * 2)

    # Counter should not increase after stop
    assert counter == 4



import asyncio
import os
import platform
import socket
from collections import Counter, defaultdict, deque
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Any, Deque, Tuple
from collections.abc import MutableMapping

import pytest

import aiomisc
from aiomisc import bind_socket
from aiomisc.service import UDPServer, sdwatchdog

pytestmark = pytest.mark.skipif(
    platform.system() == "Windows", reason="Unix only tests"
)


def test_sdwatchdog_service(event_loop):
    with TemporaryDirectory(dir="/tmp") as tmp_dir:
        tmp_path = Path(tmp_dir)
        sock_path = str(tmp_path / "notify.sock")

        packets: deque[tuple[Any, ...]] = deque()

        class FakeSystemd(UDPServer):
            async def handle_datagram(
                self, data: bytes, addr: tuple[Any, ...]
            ) -> None:
                key: str
                value: str
                key, value = data.decode().split("=", 1)
                packets.append(((key, value), addr))

        with bind_socket(
            socket.AF_UNIX,
            socket.SOCK_DGRAM,
            address=sock_path,
            reuse_port=False,
        ) as sock:
            try:
                os.environ["NOTIFY_SOCKET"] = sock_path
                os.environ["WATCHDOG_USEC"] = "100000"

                service = sdwatchdog.SDWatchdogService(
                    watchdog_interval=sdwatchdog._get_watchdog_interval()
                )

                assert service.watchdog_interval == 0.1

                with aiomisc.entrypoint(
                    FakeSystemd(sock=sock), service, loop=event_loop
                ):
                    event_loop.run_until_complete(asyncio.sleep(1))
            finally:
                for key in ("NOTIFY_SOCKET", "WATCHDOG_USEC"):
                    os.environ.pop(key)

    assert packets
    messages_count: MutableMapping[str, int] = Counter()
    messages = defaultdict(set)

    for (key, value), sender in packets:
        assert key
        assert value
        messages_count[key] += 1
        messages[key].add(value)

    assert 5 < messages_count["WATCHDOG"] < 25
    assert messages_count["STATUS"] == 2
    assert messages_count["WATCHDOG_USEC"] == 1
    assert tuple(messages["WATCHDOG_USEC"])[0] == "100000"



import asyncio
import logging
import math
import platform
import time
from asyncio import Event, wait
from contextvars import ContextVar
from typing import Any, List
from collections.abc import Sequence

import pytest

from aiomisc.aggregate import Arg, ResultNotSetError, aggregate, aggregate_async

log = logging.getLogger(__name__)

pytestmark = pytest.mark.skipif(
    platform.system() == "Windows",
    reason="Skip flapping tests on windows because it "
    "system timer hasn't enough resolution",
)


@pytest.fixture(scope="session")
def leeway() -> float:
    async def func() -> float:
        loop = asyncio.get_event_loop()
        t = loop.time()
        await asyncio.sleep(0)
        return loop.time() - t

    async def run() -> Sequence[float]:
        tasks = [asyncio.create_task(func()) for _ in range(100)]
        return await asyncio.gather(*tasks)

    ts: Sequence[float] = asyncio.run(run())
    estimated = max(ts) * 5
    default = 0.1
    result = max(estimated, default)

    if estimated > default:
        log.warning("Slow system: leeway increased to %.2f s", result)

    return result


async def test_invalid_func():
    with pytest.raises(ValueError) as excinfo:

        @aggregate(10)
        async def pho(a, b=1):
            pass

    assert str(excinfo.value) == (
        "Function must accept variadic positional arguments"
    )


@pytest.mark.parametrize("leeway_ms", (-1.0, 0.0))
async def test_invalid_leeway(leeway_ms):
    with pytest.raises(ValueError) as excinfo:

        @aggregate(leeway_ms)
        async def pho(*args):
            pass

    assert str(excinfo.value) == "leeway_ms must be positive float"


@pytest.mark.parametrize("max_count", (-1, 0))
async def test_invalid_max_count(max_count):
    with pytest.raises(ValueError) as excinfo:

        @aggregate(10, max_count)
        async def pho(*args):
            pass

    assert str(excinfo.value) == "max_count must be positive int or None"


async def test_error(event_loop, leeway):
    event = Event()

    @aggregate(leeway * 1000)
    async def pow(*args: float, power: float = 2) -> Any:
        event.set()
        raise ValueError

    async def pho(num: int):
        return await pow(float(num))

    tasks = []
    for i in range(10):
        tasks.append(event_loop.create_task(pho(i)))

    await event.wait()

    await wait(tasks)
    for task in tasks:
        assert task.done()
        assert isinstance(task.exception(), ValueError)


async def test_leeway_ok(event_loop, leeway):
    t_exec: float = 0.0
    event: Event = Event()

    @aggregate(leeway * 1000)
    async def pow(*args: float, power: float = 2) -> list[float]:
        nonlocal t_exec
        t_exec = time.time()
        event.set()

        return [math.pow(num, power) for num in args]

    tasks = []
    for i in range(9):
        tasks.append(event_loop.create_task(pow(i)))

    t = time.time()

    await asyncio.sleep(leeway * 0.1)
    assert all(not task.done() for task in tasks)

    await event.wait()
    elapsed = t_exec - t
    assert leeway * 0.9 < elapsed < leeway * 2

    await wait(tasks)
    for i, task in enumerate(tasks):
        assert task.done()
        assert task.result() == math.pow(i, 2)


async def test_max_count(event_loop, leeway):
    t_exec: float = 0.0
    event = Event()
    max_count = 5

    @aggregate(leeway * 1000, max_count)
    async def pow(*args: float, power: float = 2) -> list[float]:
        nonlocal t_exec
        t_exec = time.time()
        event.set()

        return [math.pow(num, power) for num in args]

    tasks = []
    for i in range(5):
        tasks.append(event_loop.create_task(pow(i)))

    t = time.time()

    await event.wait()
    elapsed = t_exec - t
    assert 0 < elapsed < leeway * 2

    await wait(tasks)
    for i, task in enumerate(tasks):
        assert tasks[i].done()
        assert task.result() == math.pow(i, 2)


async def test_max_count_multiple_batches(event_loop, leeway):
    t_exec: float = 0.0
    event = Event()
    max_count = 5

    @aggregate(leeway * 1000, max_count)
    async def pow(*args: float, power: float = 2) -> list[float]:
        nonlocal t_exec
        t_exec = time.time()
        event.set()

        return [math.pow(num, power) for num in args]

    tasks = []
    for i in range(9):
        tasks.append(event_loop.create_task(pow(i)))

    t = time.time()

    # Wait for the first batch
    await event.wait()
    event.clear()
    elapsed = t_exec - t
    assert 0 < elapsed < leeway

    await wait(tasks[:5])
    for i in range(5):
        assert tasks[i].done()
    for i in range(5, 9):
        assert not tasks[i].done()

    tasks.append(event_loop.create_task(pow(9)))

    # Wait for the second batch
    await event.wait()
    await wait(tasks[5:])
    for i, task in enumerate(tasks):
        assert task.done()
        assert task.result() == math.pow(i, 2)


async def test_leeway_cancel(event_loop, leeway):
    t_exec: float = 0.0
    delay_exec = 0.1
    event = Event()
    executions = 0
    arg: ContextVar = ContextVar("arg")
    tasks: list[asyncio.Task] = []
    executing_task: asyncio.Task

    @aggregate(leeway * 1000)
    async def pow(*args: float, power: float = 2) -> list[float]:
        nonlocal executions, executing_task, t_exec, delay_exec
        t_exec = time.time()
        executions += 1
        executing_task = tasks[arg.get()]
        event.set()

        await asyncio.sleep(delay_exec)
        return [math.pow(num, power) for num in args]

    async def pho(num: int):
        arg.set(num)
        return await pow(float(num))

    for i in range(9):
        tasks.append(event_loop.create_task(pho(i)))

    t = time.time()

    # Execution must have started
    await event.wait()
    event.clear()
    elapsed = t_exec - t
    assert leeway * 0.9 < elapsed < leeway * 2
    assert executions == 1
    first_executing_task: asyncio.Task = executing_task
    first_executing_task.cancel()

    # Another task must have tried to execute
    await event.wait()
    assert executions == 2
    assert first_executing_task.cancelled()
    assert all(
        not task.done() for task in tasks if task is not first_executing_task
    )

    # Must have finished
    await wait(tasks)
    assert first_executing_task.cancelled()
    for i, task in enumerate(tasks):
        if task is first_executing_task:
            continue
        assert task.done()
        assert task.result() == math.pow(i, 2)


async def test_max_count_cancel(event_loop):
    t_exec: float = 0.0
    delay_exec = 0.1
    event = Event()
    executions = 0
    leeway = 100
    max_count = 5
    arg: ContextVar = ContextVar("arg")
    tasks: list[asyncio.Task] = []
    executing_task: asyncio.Task

    @aggregate(leeway * 1000, max_count)
    async def pow(*args: float, power: float = 2) -> list[float]:
        nonlocal executions, executing_task, t_exec, delay_exec
        t_exec = time.time()
        executions += 1
        executing_task = tasks[arg.get()]
        event.set()

        await asyncio.sleep(delay_exec)
        return [math.pow(num, power) for num in args]

    async def pho(num: int):
        arg.set(num)
        return await pow(float(num))

    tasks = []
    for i in range(5):
        tasks.append(event_loop.create_task(pho(i)))

    t = time.time()

    # Execution must have started
    await event.wait()
    event.clear()
    elapsed = t_exec - t
    assert 0 < elapsed < leeway
    assert all(not task.done() for task in tasks)
    assert executions == 1
    first_executing_task = executing_task
    first_executing_task.cancel()

    # Another task must have tried to execute
    await event.wait()
    assert executions == 2
    assert first_executing_task.cancelled()
    assert all(
        not task.done() for task in tasks if task is not first_executing_task
    )

    # Must have finished
    await wait(tasks)
    assert first_executing_task.cancelled()
    for i, task in enumerate(tasks):
        if task is first_executing_task:
            continue
        assert task.done()
        assert task.result() == math.pow(i, 2)


async def test_max_count_multiple_batches_cancel(event_loop, leeway):
    delay_exec = 0.1
    event = Event()
    executions = 0
    max_count = 5
    arg: ContextVar = ContextVar("arg")
    tasks: list[asyncio.Task] = []
    executing_task: asyncio.Task

    @aggregate(leeway * 1000, max_count)
    async def pow(*args: float, power: float = 2) -> list[float]:
        nonlocal executions, executing_task, delay_exec
        executions += 1
        executing_task = tasks[arg.get()]
        event.set()

        await asyncio.sleep(delay_exec)
        return [math.pow(num, power) for num in args]

    async def pho(num: int):
        arg.set(num)
        return await pow(float(num))

    tasks = []
    for i in range(9):
        tasks.append(event_loop.create_task(pho(i)))

    # Execution of the first batch must have started
    await event.wait()
    event.clear()
    assert all(not task.done() for task in tasks)
    assert executions == 1
    first_executing_task = executing_task
    first_executing_task.cancel()

    # Another task must have tried to execute
    await event.wait()
    event.clear()
    assert executions == 2
    assert first_executing_task.cancelled()
    assert all(
        not task.done() for task in tasks if task is not first_executing_task
    )

    await wait(tasks[:5])
    # First batch must have finished
    assert first_executing_task.cancelled()
    for i, task in enumerate(tasks[:5]):
        if task is first_executing_task:
            continue
        assert task.done()
        assert task.result() == math.pow(i, 2)

    tasks.append(event_loop.create_task(pho(9)))
    # Second batch must have started execution
    await event.wait()
    assert all(not task.done() for task in tasks[5:])
    assert executions == 3

    # Second batch mast have finished
    await wait(tasks[5:])
    for i, task in enumerate(tasks[5:], start=5):
        assert task.done()
        assert task.result() == math.pow(i, 2)


async def test_low_level_sloppy(event_loop, leeway):
    max_count = 2

    @aggregate_async(leeway * 1000, max_count=max_count)
    async def pho(*args: Arg):
        for arg in args:
            if arg.value:
                arg.future.set_result(True)

    task1 = event_loop.create_task(pho(True))
    task2 = event_loop.create_task(pho(False))
    await wait([task1, task2])

    assert task1.done()
    assert await task1
    assert task2.done()
    assert isinstance(task2.exception(), ResultNotSetError)


async def test_low_level_ok(event_loop, leeway):
    @aggregate_async(leeway * 1000)
    async def pow(*args: Arg, power: float = 2):
        for arg in args:
            arg.future.set_result(math.pow(arg.value, power))

    tasks = []
    for i in range(5):
        tasks.append(event_loop.create_task(pow(i)))

    await wait(tasks)
    for i, task in enumerate(tasks):
        assert tasks[i].done()
        assert task.result() == math.pow(i, 2)


async def test_low_level_error(event_loop, leeway):
    @aggregate_async(leeway * 1000)
    async def pho(*args: Arg):
        for arg in args:
            if arg.value:
                arg.future.set_result(True)
            else:
                arg.future.set_exception(ValueError)

    task1 = event_loop.create_task(pho(True))
    task2 = event_loop.create_task(pho(False))
    await wait([task1, task2])

    assert task1.done()
    assert task1.result()
    assert task2.done()
    assert isinstance(task2.exception(), ValueError)



import asyncio
from unittest.mock import patch

import pytest

import aiomisc
from aiomisc.cron import CronCallback
from aiomisc.service.cron import CronService

pytestmark = pytest.mark.catch_loop_exceptions


# Fixed delay for deterministic testing
TICK_DELAY = 0.1


def mock_get_next(*args):
    """Return a fixed delay for deterministic testing."""
    return TICK_DELAY


@patch.object(CronCallback, "get_next", mock_get_next)
def test_cron():
    counter = 0
    condition: asyncio.Condition

    async def callback():
        nonlocal counter
        async with condition:
            counter += 1
            condition.notify_all()

    svc = CronService()

    svc.register(callback, spec="* * * * * *")

    async def assert_counter():
        nonlocal counter, svc

        counter = 0
        async with condition:
            await asyncio.wait_for(
                condition.wait_for(lambda: counter == 1), timeout=2
            )

        await svc.stop()

        await asyncio.sleep(TICK_DELAY * 2)
        async with condition:
            await asyncio.wait_for(
                condition.wait_for(lambda: counter == 1), timeout=2
            )

        assert counter == 1

    with aiomisc.entrypoint(svc) as loop:
        condition = asyncio.Condition()

        loop.run_until_complete(asyncio.wait_for(assert_counter(), timeout=10))


@patch.object(CronCallback, "get_next", mock_get_next)
def test_register():
    counter = 0
    condition: asyncio.Condition

    async def callback():
        nonlocal counter
        async with condition:
            counter += 1
            condition.notify_all()

    svc = CronService()

    svc.register(callback, spec="* * * * * *")
    svc.register(callback, spec="* * * * * */2")  # even second
    svc.register(callback, spec="* * * * * *")

    async def assert_counter():
        nonlocal counter, svc

        counter = 0
        # With mocked timing, all 3 callbacks fire every TICK_DELAY
        # After ~2 ticks we expect 6 calls (3 callbacks * 2 ticks)
        async with condition:
            await asyncio.wait_for(
                condition.wait_for(lambda: counter >= 5), timeout=10
            )

        await svc.stop()

        final_counter = counter
        await asyncio.sleep(TICK_DELAY * 2)
        # Counter should not increase after stop
        assert counter == final_counter

    with aiomisc.entrypoint(svc) as loop:
        condition = asyncio.Condition()
        loop.run_until_complete(asyncio.wait_for(assert_counter(), timeout=10))



import asyncio
import concurrent.futures
import gc
import os
import threading
import time
import weakref
from contextlib import suppress

import pytest
from async_timeout import timeout

import aiomisc
from aiomisc import threaded, threaded_iterable
from aiomisc.iterator_wrapper import ChannelClosed, FromThreadChannel

try:
    import contextvars
except ImportError:
    contextvars = None  # type: ignore


pytestmark = pytest.mark.catch_loop_exceptions


thread_pool_implementation = (
    aiomisc.ThreadPoolExecutor,
    concurrent.futures.ThreadPoolExecutor,
)


thread_pool_ids = ("aiomisc pool", "default pool")


@pytest.fixture(params=thread_pool_implementation, ids=thread_pool_ids)
def thread_pool_executor(request):
    return request.param


@pytest.fixture(params=(aiomisc.threaded, aiomisc.threaded_separate))
def threaded_decorator(request, executor: aiomisc.ThreadPoolExecutor):
    assert executor
    return request.param


@pytest.fixture
def executor(event_loop: asyncio.AbstractEventLoop):
    thread_pool = aiomisc.ThreadPoolExecutor(8)
    event_loop.set_default_executor(thread_pool)
    try:
        yield thread_pool
    finally:
        with suppress(Exception):
            thread_pool.shutdown(wait=True)

        thread_pool.shutdown(wait=True)


async def test_from_thread_channel(threaded_decorator):
    channel = FromThreadChannel(maxsize=2)

    @threaded_decorator
    def in_thread():
        with channel:
            for i in range(10):
                channel.put(i)

    in_thread()
    result = []
    with pytest.raises(ChannelClosed):
        while True:
            result.append(await asyncio.wait_for(channel.get(), timeout=5))

    assert result == list(range(10))


async def test_from_thread_channel_wait_before(event_loop, threaded_decorator):
    channel = FromThreadChannel(maxsize=1)

    @threaded_decorator
    def in_thread():
        with channel:
            for i in range(10):
                channel.put(i)

    event_loop.call_later(0.1, in_thread)

    result = []
    with pytest.raises(ChannelClosed):
        while True:
            result.append(await asyncio.wait_for(channel.get(), timeout=5))

    assert result == list(range(10))


async def test_from_thread_channel_close(event_loop):
    channel = FromThreadChannel(maxsize=1)
    with channel:
        channel.put(1)

    with pytest.raises(ChannelClosed):
        channel.put(2)

    channel = FromThreadChannel(maxsize=1)
    task = event_loop.create_task(channel.get())

    with pytest.raises(asyncio.TimeoutError):
        await asyncio.wait_for(task, timeout=1)

    event_loop.call_soon(channel.put, 1)

    assert await channel.get() == 1


async def test_future_gc(thread_pool_executor, event_loop):
    thread_pool = thread_pool_executor(2)
    event = threading.Event()

    async def run():
        future = event_loop.create_future()

        cfuture = thread_pool.submit(time.sleep, 0.5)

        cfuture.add_done_callback(
            lambda *_: event_loop.call_soon_threadsafe(future.set_result, True)
        )

        weakref.finalize(cfuture, lambda *_: event.set())
        await future

    await run()

    gc.collect()
    event.wait(1)

    assert event.is_set()


async def test_threaded(threaded_decorator, timer):
    sleep = threaded_decorator(time.sleep)

    with timer(1):
        await asyncio.wait_for(
            asyncio.gather(sleep(1), sleep(1), sleep(1), sleep(1), sleep(1)),
            timeout=5,
        )


async def test_threaded_exc(threaded_decorator):
    @threaded_decorator
    def worker():
        raise Exception

    number = 90

    done, _ = await asyncio.wait([worker() for _ in range(number)], timeout=1)

    for task in done:
        with pytest.raises(Exception):
            task.result()


async def test_future_already_done(executor: aiomisc.ThreadPoolExecutor):
    futures = []

    async with timeout(10):
        for _ in range(10):
            futures.append(executor.submit(time.sleep, 0.1))

        for future in futures:
            future.set_exception(asyncio.CancelledError())

        await asyncio.gather(*futures, return_exceptions=True)


async def test_future_when_pool_shutting_down(executor):
    futures = []

    async with timeout(10):
        for _ in range(10):
            futures.append(executor.submit(time.sleep, 0.1))

        executor.shutdown(wait=False)

        done, _ = await asyncio.wait(futures)

        for task in done:
            with pytest.raises(RuntimeError):
                task.result()


async def test_failed_future_already_done(executor):
    futures = []

    def exc():
        time.sleep(0.1)
        raise Exception

    async with timeout(10):
        for _ in range(10):
            futures.append(executor.submit(exc))

        for future in futures:
            future.set_exception(asyncio.CancelledError())

        await asyncio.gather(*futures, return_exceptions=True)


async def test_cancel(executor, event_loop, timer):
    sleep = aiomisc.threaded(time.sleep)

    async with timeout(2):
        with timer(1, dispersion=2):
            tasks = [asyncio.ensure_future(sleep(1)) for _ in range(1000)]

            await asyncio.sleep(1)

            for task in tasks:
                task.cancel()

    executor.shutdown(wait=True)


async def test_simple(threaded_decorator, event_loop, timer):
    sleep = threaded_decorator(time.sleep)

    async with timeout(2):
        with timer(1):
            await asyncio.gather(sleep(1), sleep(1), sleep(1), sleep(1))


gen_decos = (aiomisc.threaded_iterable, aiomisc.threaded_iterable_separate)


@pytest.fixture(params=gen_decos)
def iterator_decorator(request):
    return request.param


async def test_threaded_generator(event_loop, timer):
    @aiomisc.threaded
    def arange(*args):
        return (yield from range(*args))

    async with timeout(10):
        count = 10

        result = []
        agen = arange(count)
        async for item in agen:
            result.append(item)

        assert result == list(range(count))


async def test_threaded_generator_max_size(
    iterator_decorator, event_loop, timer
):
    @iterator_decorator(max_size=1)
    def arange(*args):
        return (yield from range(*args))

    async with timeout(2):
        count = 10

        result = []
        agen = arange(count)
        async for item in agen:
            result.append(item)

        assert result == list(range(count))


async def test_threaded_generator_exception(
    iterator_decorator, event_loop, timer
):
    @iterator_decorator
    def arange(*args):
        yield from range(*args)
        raise ZeroDivisionError

    async with timeout(2):
        count = 10

        result = []
        agen = arange(count)

        with pytest.raises(ZeroDivisionError):
            async for item in agen:
                result.append(item)

        assert result == list(range(count))


async def test_threaded_generator_close(iterator_decorator, event_loop, timer):
    stopped = False

    @iterator_decorator(max_size=2)
    def noise():
        nonlocal stopped

        try:
            while True:
                yield os.urandom(32)
        finally:
            stopped = True

    async with timeout(2):
        counter = 0

        async with noise() as gen:
            async for _ in gen:
                counter += 1
                if counter > 9:
                    break

        wait_counter = 0
        while not stopped and wait_counter < 5:
            await asyncio.sleep(1)
            wait_counter += 1

        assert stopped


async def test_threaded_generator_close_cm(
    iterator_decorator, event_loop, timer
):
    stopped = threading.Event()

    @iterator_decorator(max_size=1)
    def noise():
        nonlocal stopped

        try:
            while True:
                yield os.urandom(32)
        finally:
            stopped.set()

    async with timeout(2):
        async with noise() as gen:
            counter = 0
            async for _ in gen:
                counter += 1
                if counter > 9:
                    break

        stopped.wait(timeout=5)
        assert stopped.is_set()


async def test_threaded_generator_close_break(
    iterator_decorator, event_loop, timer
):
    stopped = threading.Event()

    @iterator_decorator(max_size=1)
    def noise():
        nonlocal stopped

        try:
            while True:
                yield os.urandom(32)
        finally:
            stopped.set()

    async with timeout(2):
        counter = 0
        async for _ in noise():
            counter += 1
            if counter > 9:
                break

        stopped.wait(timeout=5)
        assert stopped.is_set()


async def test_threaded_generator_non_generator_raises(
    iterator_decorator, event_loop, timer
):
    @iterator_decorator()
    def errored():
        raise RuntimeError("Aaaaaaaa")

    async with timeout(2):
        with pytest.raises(RuntimeError):
            async for _ in errored():
                pass


async def test_threaded_generator_func_raises(
    iterator_decorator, event_loop, timer
):
    @iterator_decorator
    def errored(val):
        if val:
            raise RuntimeError("Aaaaaaaa")

        yield

    async with timeout(2):
        with pytest.raises(RuntimeError):
            async for _ in errored(True):
                pass


@pytest.mark.skipif(contextvars is None, reason="no contextvars support")
async def test_context_vars(threaded_decorator, event_loop):
    ctx_var = contextvars.ContextVar("test")  # type: ignore

    @threaded_decorator
    def test(arg):
        value = ctx_var.get()
        assert value == arg * arg

    futures = []

    for i in range(8):
        ctx_var.set(i * i)
        futures.append(test(i))

    await asyncio.gather(*futures)


async def test_wait_coroutine_sync(threaded_decorator, event_loop):
    result = 0

    async def coro():
        nonlocal result
        await asyncio.sleep(1)
        result = 1

    @threaded_decorator
    def test():
        aiomisc.sync_wait_coroutine(event_loop, coro)

    await test()
    assert result == 1


async def test_wait_coroutine_sync_current_loop(threaded_decorator):
    result = 0

    async def coro():
        nonlocal result
        await asyncio.sleep(1)
        result = 1

    @threaded_decorator
    def test():
        aiomisc.wait_coroutine(coro())

    await test()
    assert result == 1


async def test_wait_awaitable(threaded_decorator):
    result = 0

    @threaded_decorator
    def in_thread():
        nonlocal result
        result += 1

    @threaded_decorator
    def test():
        aiomisc.sync_await(in_thread)

    await test()
    assert result == 1


async def test_wait_coroutine_sync_exc(threaded_decorator, event_loop):
    result = 0

    async def coro():
        nonlocal result
        await asyncio.sleep(1)
        result = 1
        raise RuntimeError("Test")

    @threaded_decorator
    def test():
        aiomisc.sync_wait_coroutine(event_loop, coro)

    with pytest.raises(RuntimeError):
        await test()

    assert result == 1


async def test_wait_coroutine_sync_exc_noloop(threaded_decorator, event_loop):
    result = 0

    async def coro():
        nonlocal result
        await asyncio.sleep(1)
        result = 1
        raise RuntimeError("Test")

    @threaded_decorator
    def test():
        aiomisc.sync_wait_coroutine(None, coro)

    with pytest.raises(RuntimeError):
        await test()

    assert result == 1


def test_task_channel():
    channel = aiomisc.thread_pool.TaskChannel()
    events = []

    def consumer(event: threading.Event):
        try:
            while True:
                channel.get()(no_return=True)
        except aiomisc.thread_pool.TaskChannelCloseException:
            event.set()

    for _ in range(500):
        event = threading.Event()
        threading.Thread(target=consumer, args=(event,)).start()
        events.append(event)

    loop = asyncio.get_event_loop()
    future = loop.create_future()
    item = aiomisc.thread_pool.WorkItem(
        func=lambda x: None,
        loop=loop,
        future=future,
        statistic=aiomisc.thread_pool.ThreadPoolStatistic(),
    )

    for _ in range(10000):
        channel.put_nowait(item)

    channel.close()

    for event in events:
        event.wait(timeout=1)


async def test_threaded_class_func():
    @threaded
    def foo():
        return 42

    assert foo.sync_call() == 42
    assert await foo() == 42
    assert await foo.async_call() == 42


async def test_threaded_class_method():
    class TestClass:
        @threaded
        def foo(self):
            return 42

    instance = TestClass()
    assert instance.foo.sync_call() == 42
    assert await instance.foo() == 42
    assert await instance.foo.async_call() == 42


async def test_threaded_class_staticmethod():
    class TestClass:
        @threaded
        @staticmethod
        def foo():
            return 42

    instance = TestClass()
    assert instance.foo.sync_call() == 42
    assert await instance.foo() == 42
    assert await instance.foo.async_call() == 42


async def test_threaded_class_classmethod():
    class TestClass:
        @threaded
        @classmethod
        def foo(cls):
            return 42

    instance = TestClass()
    assert instance.foo.sync_call() == 42
    assert await instance.foo() == 42
    assert await instance.foo.async_call() == 42


async def test_threaded_iterator_class_func():
    @threaded_iterable
    def foo():
        yield 42

    assert list(foo.sync_call()) == [42]
    assert [x async for x in foo()] == [42]
    assert [x async for x in foo.async_call()] == [42]


async def test_threaded_iterator_class_method():
    class TestClass:
        @threaded_iterable
        def foo(self):
            yield 42

    instance = TestClass()
    assert list(instance.foo.sync_call()) == [42]
    assert [x async for x in instance.foo()] == [42]
    assert [x async for x in instance.foo.async_call()] == [42]


async def test_threaded_iterator_class_staticmethod():
    class TestClass:
        @threaded_iterable
        @staticmethod
        def foo():
            yield 42

    instance = TestClass()
    assert list(instance.foo.sync_call()) == [42]
    assert [x async for x in instance.foo()] == [42]
    assert [x async for x in instance.foo.async_call()] == [42]


async def test_threaded_iterator_class_classmethod():
    class TestClass:
        @threaded_iterable
        @classmethod
        def foo(cls):
            yield 42

    instance = TestClass()
    assert list(instance.foo.sync_call()) == [42]
    assert [x async for x in instance.foo()] == [42]
    assert [x async for x in instance.foo.async_call()] == [42]


async def test_threaded_generator_starts_in_aenter():
    """Test that the generator starts running when entering async context."""
    generator_started = threading.Event()

    @aiomisc.threaded_iterable(max_size=2)
    def gen():
        generator_started.set()
        yield 1
        yield 2
        yield 3

    async with timeout(2):
        async with gen() as iterator:
            # The generator should have started by now
            generator_started.wait(timeout=1)
            assert generator_started.is_set()

            # Verify we can still iterate
            result = [x async for x in iterator]
            assert result == [1, 2, 3]



import logging
import struct
from asyncio import (
    Future,
    IncompleteReadError,
    Lock,
    Queue,
    StreamReader,
    StreamWriter,
    iscoroutinefunction,
)
from io import BytesIO
from typing import Any, Dict, Optional
from collections.abc import AsyncIterable, Callable, Mapping, MutableMapping

import msgspec

from aiomisc import cancel_tasks
from aiomisc.service.base import TaskStoreBase

log = logging.getLogger()


class Request(msgspec.Struct):
    id: int
    method: str
    params: dict[str, Any]


class Error(msgspec.Struct):
    type: str
    args: Any


class Response(msgspec.Struct):
    id: int
    result: Any | None = None
    error: Error | None = None


class Payload(msgspec.Struct):
    request: Request | None = None
    response: Response | None = None


class Protocol:
    PACKET_HEADER = struct.Struct(">I")
    reader: StreamReader
    writer: StreamWriter

    def __init__(self, reader: StreamReader, writer: StreamWriter):
        self.reader = reader
        self.writer = writer
        self.encoder = msgspec.msgpack.Encoder()
        self.decoder = msgspec.msgpack.Decoder(Payload)
        self.write_lock = Lock()

    def __aiter__(self) -> AsyncIterable[Payload]:
        return self

    async def __anext__(self) -> Payload:
        if self.reader.at_eof():
            raise StopAsyncIteration()

        try:
            payload_size: int = self.PACKET_HEADER.unpack(
                await self.reader.readexactly(self.PACKET_HEADER.size)
            )[0]
            payload_bytes: bytes = await self.reader.readexactly(payload_size)
            return self.decoder.decode(payload_bytes)
        except IncompleteReadError:
            raise StopAsyncIteration()

    async def send(self, payload: Payload) -> None:
        payload_bytes: bytes = self.encoder.encode(payload)
        with BytesIO() as fp:
            fp.write(self.PACKET_HEADER.pack(len(payload_bytes)))
            fp.write(payload_bytes)
            packet = fp.getvalue()

        async with self.write_lock:
            self.writer.write(packet)


class RPCBase(TaskStoreBase):
    __required__ = ("handlers",)

    _futures: MutableMapping[int, Future]
    _serial: int
    _request_queue: Queue
    handlers: Mapping[str, Callable]

    async def start(self) -> None:
        self._futures = {}
        self._serial = 0
        self._request_queue = Queue()

    def _on_response(self, response: Response) -> None:
        future = self._futures.pop(response.id, None)

        if future.done():
            return

        if response.error is not None:
            future.set_exception(
                Exception(response.error.type, *response.error.args)
            )
            return
        future.set_result(response.result)

    async def execute(self, protocol: Protocol, request: Request):
        try:
            func = self.handlers[request.method]
            if iscoroutinefunction(func):
                result = await func(**request.params)
            else:
                result = func(**request.params)
            payload = Payload(response=Response(id=request.id, result=result))
        except Exception as e:
            log.exception("Execution error:")
            payload = Payload(
                response=Response(
                    id=request.id, error=Error(type=str(type(e)), args=e.args)
                )
            )
        await protocol.send(payload)

    async def _request_writer(self, protocol: Protocol) -> None:
        while True:
            payload: Payload = await self._request_queue.get()
            await protocol.send(payload)

    def __call__(self, method, **kwargs):
        self._serial += 1
        future = self.loop.create_future()
        self._futures[self._serial] = future
        self._request_queue.put_nowait(
            Payload(
                request=Request(id=self._serial, method=method, params=kwargs)
            )
        )
        return self._futures[self._serial]

    async def communicate(self, reader: StreamReader, writer: StreamWriter):
        protocol = Protocol(reader, writer)
        request_writer_task = self.create_task(self._request_writer(protocol))

        client_host, client_port = writer.get_extra_info("peername")[:2]
        if ":" in client_host:
            client_host = f"[{client_host}]"

        log.info(
            "Start communication with tcp://%s:%d", client_host, client_port
        )

        # noinspection PyBroadException
        try:
            async for payload in protocol:
                if payload.request is not None:
                    self.create_task(self.execute(protocol, payload.request))
                if payload.response is not None:
                    self._on_response(payload.response)
        except Exception:
            log.exception(
                "Error when communication tcp://%s:%d/",
                client_host,
                client_port,
            )
        finally:
            if writer.can_write_eof():
                writer.write_eof()
            writer.close()
            log.info(
                "Communication with tcp://%s:%d finished",
                client_host,
                client_port,
            )

            await cancel_tasks([request_writer_task])
            await writer.wait_closed()



import asyncio
import operator
from functools import reduce

import pytest
from rpc.client import RPCClient
from rpc.server import RPCServer


@pytest.fixture
def handlers():
    return {
        "foo": lambda: "bar",
        "mul": lambda **a: reduce(operator.mul, a.values()),
        "div": lambda **a: reduce(operator.truediv, a.values()),
    }


@pytest.fixture
def server_port_sock(aiomisc_socket_factory):
    return aiomisc_socket_factory()


@pytest.fixture
def rpc_server(server_port_sock, handlers):
    _, sock = server_port_sock
    return RPCServer(handlers=handlers, sock=sock)


@pytest.fixture
async def rpc_client(server_port_sock, localhost, handlers) -> RPCClient:
    port, _ = server_port_sock
    return RPCClient(address=localhost, port=port, handlers=handlers)


@pytest.fixture
def services(rpc_server, rpc_client):
    return [rpc_server, rpc_client]


async def test_foo(rpc_client: RPCClient):
    assert await rpc_client("foo") == "bar"


async def test_multiply(rpc_client: RPCClient):
    assert await rpc_client("mul", a=1, b=3) == 3


async def test_division(rpc_client: RPCClient):
    with pytest.raises(Exception):
        assert await rpc_client("div", a=1, b=0)

    assert await rpc_client("div", a=10, b=5) == 2.0


async def test_many(rpc_client: RPCClient):
    calls = []
    expected = []

    for i in range(100):
        calls.append(rpc_client("div", a=i, b=5))
        calls.append(rpc_client("mul", a=i, b=5))

        expected.append(i / 5)
        expected.append(i * 5)

    results = await asyncio.gather(*calls)

    assert results == expected


async def test_two_way(rpc_client: RPCClient, rpc_server: RPCServer):
    assert await rpc_server("div", a=100, b=10) == await rpc_client(
        "div", a=100, b=10
    )



import asyncio
import logging
import socket
from abc import ABC, abstractmethod
from functools import partial
from pathlib import Path
from typing import Any

from ..utils import OptionsType, TimeoutType, awaitable, bind_socket
from .base import SimpleClient, SimpleServer

log = logging.getLogger(__name__)


class TCPServer(SimpleServer):
    PROTO_NAME = "tcp"

    def __init__(
        self,
        address: str | None = None,
        port: int | None = None,
        options: OptionsType = (),
        sock: socket.socket | None = None,
        **kwargs: Any,
    ):
        if not sock:
            if address is None or port is None:
                raise RuntimeError(
                    "You should pass socket instance or "
                    '"address" and "port" couple'
                )

            self.make_socket = partial(
                bind_socket, address=address, port=port, options=options
            )
        elif not isinstance(sock, socket.socket):
            raise ValueError("sock must be socket instance")
        else:
            self.make_socket = lambda: sock  # type: ignore

        self.socket: socket.socket | None = None

        super().__init__(**kwargs)

    @abstractmethod
    async def handle_client(
        self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter
    ) -> Any:
        raise NotImplementedError

    async def __client_handler(
        self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter
    ) -> Any:
        return await awaitable(self.handle_client)(reader, writer)

    def make_client_handler(
        self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter
    ) -> Any:
        return self.create_task(self.__client_handler(reader, writer))

    async def start(self) -> None:
        self.socket = self.make_socket()
        self.server = await asyncio.start_server(
            self.make_client_handler, sock=self.socket
        )

    async def stop(self, exc: Exception | None = None) -> None:
        await super().stop(exc)

        if self.server:
            await self.server.wait_closed()

    @property
    def __sockname(self) -> tuple | None:
        if self.socket:
            return self.socket.getsockname()
        return None

    @property
    def address(self) -> str | None:
        if self.__sockname:
            return self.__sockname[0]
        return None

    @property
    def port(self) -> int | None:
        if self.__sockname:
            return self.__sockname[1]
        return None


class TCPClient(SimpleClient, ABC):
    PROTO_NAME = "tcp"

    address: str
    port: int
    path: Path

    connect_attempts: int = 5
    connect_retry_timeout: TimeoutType = 3

    def __init__(
        self, address: str | None = None, port: int | None = None, **kwargs: Any
    ):
        if not (address and port):
            raise RuntimeError('You should pass "address" and "port" couple')
        self.address = address
        self.port = port
        super().__init__(**kwargs)

    async def connect(
        self,
    ) -> tuple[asyncio.StreamReader, asyncio.StreamWriter]:
        last_error: Exception | None = None

        for _ in range(self.connect_attempts):
            try:
                reader, writer = await self.create_task(
                    asyncio.open_connection(self.address, self.port)
                )
                log.info(
                    "Connected to %s://%s:%d",
                    self.PROTO_NAME,
                    self.address,
                    self.port,
                )
                return reader, writer
            except ConnectionError as e:
                last_error = e
                await asyncio.sleep(self.connect_retry_timeout)

        raise last_error or RuntimeError("Unknown exception occurred")

    @abstractmethod
    async def handle_connection(
        self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter
    ) -> None:
        raise NotImplementedError

    async def start(self) -> None:
        reader, writer = await self.connect()
        self.loop.call_soon(self.start_event.set)
        try:
            await self.create_task(self.handle_connection(reader, writer))
        except Exception:
            log.exception("Error when handle TCP connection")
        else:
            log.info("Client %r communication has been ended", self)


class RobustTCPClient(TCPClient, ABC):
    reconnect_timeout: TimeoutType = 5
    connected: asyncio.Event

    async def start(self) -> None:
        self.connected = asyncio.Event()

        while True:
            try:
                reader, writer = await self.connect()
                self.connected.set()

                if not self.start_event.is_set():
                    self.start_event.set()

                await self.create_task(self.handle_connection(reader, writer))
                self.connected.clear()
            except asyncio.CancelledError:
                raise
            except Exception:
                log.exception(
                    "Error when handle TCP connection, "
                    "reconnecting after %.3f seconds",
                    self.reconnect_timeout,
                )
            finally:
                if not self.connected.is_set():
                    return
                await asyncio.sleep(self.reconnect_timeout)



import asyncio

import pytest

import aiomisc

pytestmark = pytest.mark.catch_loop_exceptions


async def test_simple(event_loop):
    @aiomisc.timeout(0)
    async def test():
        await asyncio.sleep(0.05)

    with pytest.raises(asyncio.TimeoutError):
        await test()


async def test_already_done_2(event_loop):
    @aiomisc.timeout(0.5)
    async def test(sec):
        await asyncio.sleep(sec)

    task = event_loop.create_task(test(10))
    task.cancel()

    with pytest.raises(asyncio.CancelledError):
        await task


async def test_non_coroutine(event_loop):
    with pytest.raises(TypeError):

        @aiomisc.timeout(0)
        def test():
            return



import asyncio
import io
import socket
from threading import Event

from aiomisc import threaded
from aiomisc_worker import protocol


def test_io_protocol(tmp_path):
    with open(tmp_path / "test.bin", "wb") as fp:
        proto = protocol.FileIOProtocol(fp)
        proto.send(("foo", "bar", print))

    with open(tmp_path / "test.bin", "rb") as fp:
        proto = protocol.FileIOProtocol(fp)
        assert proto.receive() == ("foo", "bar", print)


async def test_socket_protocol(localhost, aiomisc_unused_port):
    server_start_event = Event()
    server_data = []
    client_data = []

    @threaded
    def server():
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock:
            sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            sock.bind((localhost, aiomisc_unused_port))
            sock.listen(1)
            server_start_event.set()
            cs, addr = sock.accept()
            proto = protocol.SocketIOProtocol(cs)
            server_data.append(proto.receive())
            proto.send({"foo": "bar"})

    @threaded
    def client():
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock:
            server_start_event.wait(5)
            sock.connect((localhost, aiomisc_unused_port))

            proto = protocol.SocketIOProtocol(sock)
            proto.send({"foo": "bar"})
            client_data.append(proto.receive())

    await asyncio.gather(server(), client())
    assert server_data == [{"foo": "bar"}]
    assert client_data == [{"foo": "bar"}]
    assert server_data == client_data


async def test_socket_protocol_partial_read(localhost, aiomisc_unused_port):
    server_start_event = Event()
    server_data = []
    client_data = []

    @threaded
    def server():
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock:
            sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            sock.bind((localhost, aiomisc_unused_port))
            sock.listen(1)
            server_start_event.set()
            cs, addr = sock.accept()
            proto = protocol.SocketIOProtocol(cs)
            data = proto._pack(list(range(10_000)))
            with io.BytesIO(data) as fp:
                chunk = fp.read(1024)
                while chunk:
                    proto.sock.send(chunk)
                    chunk = fp.read(1024)
            server_data.append(proto.receive())

    @threaded
    def client():
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock:
            server_start_event.wait(5)
            sock.connect((localhost, aiomisc_unused_port))

            proto = protocol.SocketIOProtocol(sock)
            data = proto.receive()
            proto.send(len(data))
            client_data.append(data)

    await asyncio.gather(server(), client())
    assert server_data == [10_000]
    assert client_data == [list(range(10_000))]



import asyncio

import pytest

import aiomisc

pytestmark = pytest.mark.catch_loop_exceptions


async def test_base_class(event_loop: asyncio.AbstractEventLoop):
    with pytest.raises(TypeError):
        aiomisc.PoolBase()  # type: ignore


class SimplePool(aiomisc.PoolBase):
    async def _create_instance(self) -> asyncio.Future:
        assert self._loop is not None
        return self._loop.create_future()

    async def _destroy_instance(self, instance):
        if instance.done():
            return

        instance.set_result(True)

    async def _check_instance(self, instance):
        return not instance.done()


async def test_simple_pool_no_reuse_context_manager(event_loop):
    size = 5
    recycle = 1

    pool = SimplePool(maxsize=size, recycle=recycle)

    cm = pool.acquire()

    async with cm as future:
        assert not future.done()

    with pytest.raises(RuntimeError):
        async with cm:
            pass

    await pool.close()


async def test_simple_pool_recycle(event_loop):
    size = 5
    recycle = 1

    pool = SimplePool(maxsize=size, recycle=recycle)

    futures = []

    def on_fail(future):
        if future.done():
            return

        future.set_exception(TimeoutError)

    async def run():
        async with pool.acquire() as future:
            assert not future.done()
            futures.append(future)
            event_loop.call_later(5 * recycle, on_fail, future)

    await asyncio.gather(*[run() for _ in range(size)])
    assert len(pool) == size

    await asyncio.sleep(2 * recycle)

    async def run_blank():
        async with pool.acquire():
            pass

    await asyncio.gather(*[run_blank() for _ in range(size)])

    await asyncio.gather(*futures)
    await pool.close()


async def test_simple_pool_check_before(event_loop):
    size = 5
    pool = SimplePool(maxsize=size, recycle=None)

    futures = set()

    for _ in range(size * 2):
        async with pool.acquire() as future:
            assert not future.done()
            futures.add(future)
            future.set_result(True)

    assert len(futures) == size * 2

    await asyncio.wait_for(asyncio.gather(*futures), timeout=1)

    await pool.close()


async def test_simple_pool_check_after(event_loop):
    size = 5
    pool = SimplePool(maxsize=size, recycle=None)

    futures = set()

    for _ in range(size * 2):
        # Switch context
        await asyncio.sleep(0)

        async with pool.acquire() as future:
            assert not future.done()
            futures.add(future)
            event_loop.call_soon(future.set_result, True)

    assert len(futures) == size * 2
    await asyncio.gather(*futures)
    await pool.close()


async def test_simple_pool_parallel(event_loop):
    size = 5
    pool = SimplePool(maxsize=size, recycle=None)

    async def run():
        async with pool.acquire() as future:
            assert not future.done()

    tasks = [event_loop.create_task(run()) for _ in range(1000)]

    await asyncio.gather(*tasks)

    assert len(pool) == size
    await pool.close()


async def test_simple_pool_parallel_broken_instances(event_loop):
    size = 5
    pool = SimplePool(maxsize=size, recycle=None)

    async def run():
        async with pool.acquire() as future:
            assert not future.done()
            future.set_result(True)

    tasks = [event_loop.create_task(run()) for _ in range(1000)]

    await asyncio.gather(*tasks)

    assert len(pool) == size
    await pool.close()



import asyncio
from typing import Any

import pytest

import aiomisc
from aiomisc.service.periodic import PeriodicService

pytestmark = pytest.mark.catch_loop_exceptions


def test_str_representation():
    class FooPeriodicService(PeriodicService):
        async def callback(self) -> Any:
            pass

    svc = FooPeriodicService(interval=42, delay=4815162342)
    assert str(svc) == "FooPeriodicService(interval=42,delay=4815162342)"


def test_periodic(event_loop):
    counter = 0
    condition = asyncio.Condition()

    class CountPeriodicService(PeriodicService):
        async def callback(self):
            nonlocal counter
            nonlocal condition

            async with condition:
                counter += 1
                condition.notify_all()

    svc = CountPeriodicService(interval=0.1)

    async def assert_counter():
        nonlocal counter, svc

        counter = 0

        for i in (5, 10):
            async with condition:
                await asyncio.wait_for(
                    condition.wait_for(lambda: counter >= i), timeout=10
                )
            assert counter == i

    with aiomisc.entrypoint(svc, loop=event_loop) as loop:
        loop.run_until_complete(asyncio.wait_for(assert_counter(), timeout=10))


def test_delay(event_loop):
    counter = 0
    condition = asyncio.Condition()

    class CountPeriodicService(PeriodicService):
        async def callback(self):
            nonlocal counter

            async with condition:
                counter += 1
                condition.notify_all()

    svc = CountPeriodicService(interval=0.1, delay=0.5)

    async def assert_counter():
        nonlocal counter, svc

        counter = 0
        await asyncio.sleep(0.25)
        assert not counter

        async with condition:
            await asyncio.wait_for(
                condition.wait_for(lambda: counter == 5), timeout=5
            )

        await svc.stop(None)

    with aiomisc.entrypoint(svc, loop=event_loop) as loop:
        loop.run_until_complete(asyncio.wait_for(assert_counter(), timeout=10))



import asyncio
import hashlib
import hmac
import logging
import os
import signal
import socket
import sys
from types import TracebackType
from typing import Any

from . import SIGNAL, AddressType, PacketTypes
from .protocol import ADDRESS_FAMILY, SocketIOProtocol


def execute(protocol: SocketIOProtocol) -> None:
    pid = os.getpid()
    try:
        packet_type, func, args, kwargs = protocol.receive()
        logging.debug(
            "Worker PID: %d got task %r", pid, func, extra={"pid": pid}
        )
    except ConnectionError:
        raise

    if packet_type != packet_type.REQUEST:
        raise RuntimeError("Unexpected request")

    try:
        protocol.send((PacketTypes.RESULT, func(*args, **kwargs)))
    except asyncio.CancelledError:
        logging.exception("Request cancelled for %r", func, extra={"pid": pid})
        protocol.send((PacketTypes.CANCELLED, asyncio.CancelledError))
    except (ConnectionError, ConnectionResetError):
        logging.warning(
            "IPC connection error, worker PID: %d exiting.",
            pid,
            extra={"pid": pid},
        )
        raise
    except Exception as e:
        logging.exception("Exception when processing request")
        protocol.send((PacketTypes.EXCEPTION, e))


def on_cancel_signal(*_: Any) -> None:
    raise asyncio.CancelledError


signal.signal(SIGNAL, on_cancel_signal)


def on_exception(
    exc_type: type[BaseException] | None,
    exc_value: BaseException | None,
    exc_tb: TracebackType | None,
) -> None:
    if exc_type is None or exc_value is None or exc_tb is None:
        return
    pid = os.getpid()
    logging.exception(
        "Unhandled exception in worker PID: %d",
        pid,
        exc_info=(exc_type, exc_value, exc_tb),
        extra={"pid": pid},
    )


def worker(address: AddressType, cookie: bytes, worker_id: bytes) -> None:
    sys.excepthook = on_exception

    with socket.socket(ADDRESS_FAMILY, socket.SOCK_STREAM) as sock:
        try:
            sock.connect(address)
            proto = SocketIOProtocol(sock)
            proto.send(
                (
                    PacketTypes.AUTH,
                    worker_id,
                    hmac.HMAC(
                        cookie, worker_id, digestmod=hashlib.sha256
                    ).digest(),
                    os.getpid(),
                )
            )

            response = proto.receive()

            if response != PacketTypes.AUTH_OK:
                raise RuntimeError(f"Failed to authorize worker on {address!r}")

            while True:
                execute(proto)
        except ConnectionError:
            return


def bad_initializer(
    address: AddressType, cookie: bytes, worker_id: bytes, exc: BaseException
) -> None:
    with socket.socket(ADDRESS_FAMILY, socket.SOCK_STREAM) as sock:
        sock.connect(address)
        proto = SocketIOProtocol(sock)
        proto.send(
            (
                PacketTypes.BAD_INITIALIZER,
                worker_id,
                hmac.HMAC(cookie, worker_id, digestmod=hashlib.sha256).digest(),
                os.getpid(),
            )
        )
        proto.send((PacketTypes.EXCEPTION, exc))



from setuptools import setup


def version_scheme(version):
    """Custom version: MAJOR.MINOR.COMMITS_SINCE_TAG"""
    if version.exact:
        return f"{version.tag}.0"
    return f"{version.tag}.{version.distance}"


def local_scheme(version):
    """No local version component."""
    return ""


setup(
    use_scm_version={
        "version_scheme": version_scheme,
        "local_scheme": local_scheme,
        "tag_regex": r"^v(?P<version>\d+\.\d+)$",
        "write_to": "aiomisc/version.py",
    }
)



import asyncio

import pytest

import aiomisc

pytestmark = pytest.mark.catch_loop_exceptions


async def test_recurring(event_loop):
    condition = asyncio.Condition()
    counter = 0

    async def task():
        nonlocal counter
        counter += 1

        async with condition:
            condition.notify_all()

    recurring = aiomisc.RecurringCallback(task)
    start_task = recurring.start(strategy=lambda _: 0, loop=event_loop)

    async with condition:
        await asyncio.wait_for(
            condition.wait_for(lambda: counter >= 5), timeout=5
        )

    await aiomisc.cancel_tasks([start_task])


async def test_long_func(event_loop):
    counter = 0
    condition = asyncio.Condition()

    async def task():
        nonlocal counter
        async with condition:
            await asyncio.sleep(0.5)
            counter += 1
            condition.notify_all()

    recurring = aiomisc.RecurringCallback(task)
    start_task = recurring.start(strategy=lambda _: 0, loop=event_loop)

    await asyncio.sleep(1.2)
    await aiomisc.cancel_tasks([start_task])

    async with condition:
        await asyncio.wait_for(
            condition.wait_for(lambda: counter >= 2), timeout=2
        )

    assert counter == 2


@aiomisc.timeout(5)
async def test_shield(event_loop):
    counter = 0
    start_event = asyncio.Event()
    stop_event = asyncio.Event()

    async def task():
        nonlocal counter
        start_event.set()
        await asyncio.sleep(0.5)
        counter += 1
        stop_event.set()

    recurring = aiomisc.RecurringCallback(task)
    start_task = recurring.start(
        strategy=lambda _: 0, loop=event_loop, shield=True
    )

    await start_event.wait()
    await aiomisc.cancel_tasks([start_task])
    await stop_event.wait()
    assert counter == 1


@aiomisc.timeout(5)
async def test_control_flow_stop(event_loop):
    stop_event = asyncio.Event()

    async def strategy(_: aiomisc.RecurringCallback):
        stop_event.set()
        raise aiomisc.StrategyStop()

    recurring = aiomisc.RecurringCallback(lambda: None)
    task = recurring.start(strategy=strategy, loop=event_loop)

    await stop_event.wait()
    await task


@aiomisc.timeout(5)
async def test_control_flow_skip(event_loop):
    start_event = asyncio.Event()
    stop_event = asyncio.Event()
    counter = 0
    strategy_counter = 0

    async def task():
        nonlocal counter
        counter += 1
        start_event.set()

    async def strategy(_: aiomisc.RecurringCallback):
        nonlocal strategy_counter

        strategy_counter += 1

        if strategy_counter == 3:
            raise aiomisc.StrategySkip(0)

        if strategy_counter == 5:
            stop_event.set()
            raise aiomisc.StrategyStop()

        return 0

    recurring = aiomisc.RecurringCallback(task)
    start_task = recurring.start(strategy=strategy, loop=event_loop)

    await start_event.wait()
    await stop_event.wait()
    await start_task

    assert counter == 3
    assert strategy_counter == 5


@aiomisc.timeout(5)
async def test_wrong_strategy(event_loop):
    counter = 0
    strategy_counter = 0

    async def task():
        nonlocal counter
        counter += 1

    async def strategy(_: aiomisc.RecurringCallback):
        nonlocal strategy_counter
        strategy_counter += 1
        return None

    recurring = aiomisc.RecurringCallback(task)
    start_task = recurring.start(strategy=strategy, loop=event_loop)

    await start_task
    assert strategy_counter == 1
    assert counter == 0



import asyncio
import operator
import socket
from functools import reduce
from typing import Any
from collections.abc import Awaitable, Callable

import pytest
from rpc.server import RPCServer


@pytest.fixture
def handlers():
    return {
        "foo": lambda: "bar",
        "mul": lambda **a: reduce(operator.mul, a.values()),
        "div": lambda **a: reduce(operator.truediv, a.values()),
    }


@pytest.fixture
def server_port_sock(aiomisc_socket_factory):
    return aiomisc_socket_factory(socket.AF_INET, socket.SOCK_DGRAM)


@pytest.fixture
async def rpc_client(localhost, loop) -> RPCServer:
    return RPCServer(handlers=None, address=localhost, port=0)


@pytest.fixture
async def rpc_server(server_port_sock, loop, handlers) -> RPCServer:
    _, sock = server_port_sock
    return RPCServer(handlers=handlers, sock=sock)


@pytest.fixture
def services(rpc_server, rpc_client):
    return [rpc_server, rpc_client]


@pytest.fixture
async def rpc_call(
    server_port_sock, localhost, rpc_client: RPCServer
) -> Callable[..., Awaitable[Any]]:
    port, _ = server_port_sock

    async def call(method, **params) -> Any:
        return await rpc_client(localhost, port, method, **params)

    return call


async def test_foo(rpc_call: Callable[..., Awaitable[Any]]):
    assert await rpc_call("foo") == "bar"


async def test_multiply(rpc_call: Callable[..., Awaitable[Any]]):
    assert await rpc_call("mul", a=1, b=3) == 3


async def test_division(rpc_call: Callable[..., Awaitable[Any]]):
    with pytest.raises(Exception):
        assert await rpc_call("div", a=1, b=0)

    assert await rpc_call("div", a=10, b=5) == 2.0


async def test_many(rpc_call: Callable[..., Awaitable[Any]]):
    calls = []
    expected = []

    for i in range(100):
        calls.append(rpc_call("div", a=i, b=5))
        calls.append(rpc_call("mul", a=i, b=5))

        expected.append(i / 5)
        expected.append(i * 5)

    results = await asyncio.gather(*calls)

    assert results == expected



import asyncio
import os
import pickle
import socket
import uuid
from asyncio import Event, get_event_loop
from asyncio.tasks import Task
from contextlib import ExitStack, suppress
from tempfile import mktemp
from types import ModuleType
from typing import Any, Optional, Set, Tuple
from unittest import mock

import aiohttp.web
import fastapi
import grpc.aio
import pytest

import aiomisc
from aiomisc import timeout
from aiomisc.entrypoint import Entrypoint, entrypoint
from aiomisc.service import TCPServer, TLSServer, UDPServer
from aiomisc.service.aiohttp import AIOHTTPService
from aiomisc.service.asgi import ASGIApplicationType, ASGIHTTPService
from aiomisc.service.grpc_server import GRPCService
from aiomisc.service.tcp import RobustTCPClient, TCPClient
from aiomisc.service.tls import RobustTLSClient, TLSClient
from aiomisc.service.uvicorn import UvicornApplication, UvicornService
from aiomisc_log import LogFormat
from aiomisc_log.enum import DateFormat, LogLevel
from tests import unix_only

try:
    import uvloop

    uvloop_loop_type = uvloop.Loop
except ImportError:
    uvloop_loop_type = None  # type: ignore


pytestmark = pytest.mark.catch_loop_exceptions


@pytest.fixture()
def unix_socket_udp():
    socket_path = mktemp(dir="/tmp", suffix=".sock")
    sock = socket.socket(socket.AF_UNIX, socket.SOCK_DGRAM)
    sock.setblocking(False)

    # Behaviour like in the bind_socket
    sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)

    try:
        sock.bind(socket_path)
        yield sock
    except Exception:
        pass
    else:
        sock.close()
    finally:
        if os.path.exists(socket_path):
            os.remove(socket_path)


@pytest.fixture()
def unix_socket_tcp():
    socket_path = mktemp(dir="/tmp", suffix=".sock")
    sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
    sock.setblocking(False)

    try:
        sock.bind(socket_path)
        yield sock
    except Exception:
        pass
    else:
        sock.close()
    finally:
        if os.path.exists(socket_path):
            os.remove(socket_path)


def test_service_class():
    with pytest.raises(TypeError):
        service = aiomisc.Service(running=False, stopped=False)  # type: ignore
        with aiomisc.entrypoint(service):
            pass


def test_simple():
    class StartingService(aiomisc.Service):
        async def start(self):
            self.running = True

    class DummyService(StartingService):
        async def stop(self, err: Exception | None = None):
            self.stopped = True

    services: tuple[StartingService, ...]
    dummy_services: tuple[DummyService, ...]

    dummy_services = (
        DummyService(running=False, stopped=False),
        DummyService(running=False, stopped=False),
    )

    with aiomisc.entrypoint(*dummy_services):
        pass

    for svc in dummy_services:
        assert svc.running
        assert svc.stopped

    dummy_services = (
        DummyService(running=False, stopped=False),
        DummyService(running=False, stopped=False),
    )

    with pytest.raises(RuntimeError):
        with aiomisc.entrypoint(*dummy_services):
            raise RuntimeError

    for svc in dummy_services:
        assert svc.running
        assert svc.stopped

    services = (StartingService(running=False), StartingService(running=False))

    with pytest.raises(RuntimeError):
        with aiomisc.entrypoint(*services):
            raise RuntimeError

    for starting_svc in services:
        assert starting_svc.running


def test_wrong_subclass():
    with pytest.raises(TypeError):

        class NoAsyncStartService(aiomisc.Service):
            def start(self):
                return True

    class MyService(aiomisc.Service):
        async def start(self):
            return

    with pytest.raises(TypeError):

        class NoAsyncStopServiceSubclass(MyService):
            def stop(self, *_) -> Any:
                return True

    class AsyncStopServiceSubclass(MyService):
        async def stop(self, *_) -> Any:
            return True


def test_required_kwargs():
    class Svc(aiomisc.Service):
        __required__ = ("foo",)

        async def start(self):
            pass

    with pytest.raises(AttributeError):
        Svc()

    assert Svc(foo="bar").foo == "bar"


def test_tcp_server():
    event = asyncio.Event()

    class TestService(TCPServer):
        DATA = []

        async def handle_client(
            self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter
        ):
            self.DATA.append(await reader.readline())
            writer.close()
            event.set()

    service = TestService("127.0.0.1", 0)

    @aiomisc.threaded
    def writer(port):
        with socket.create_connection(("127.0.0.1", port)) as sock:
            sock.send(b"hello server\n")

    with aiomisc.entrypoint(service) as loop:
        loop.run_until_complete(
            asyncio.wait_for(writer(service.port), timeout=10)
        )
        loop.run_until_complete(event.wait())

    assert TestService.DATA
    assert TestService.DATA == [b"hello server\n"]


def test_tcp_client(aiomisc_socket_factory, localhost):
    event = asyncio.Event()

    class TestService(TCPServer):
        DATA = []

        async def handle_client(
            self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter
        ):
            self.DATA.append(await reader.readline())
            event.set()

    class TestClient(TCPClient):
        async def handle_connection(
            self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter
        ) -> None:
            writer.write(b"hello server\n")
            await writer.drain()

    port, sock = aiomisc_socket_factory()
    event = asyncio.Event()
    services = [
        TestService(sock=sock),
        TestClient(address=localhost, port=port),
    ]

    async def go():
        await event.wait()

    with aiomisc.entrypoint(*services) as loop:
        loop.run_until_complete(asyncio.wait_for(go(), timeout=10))
        loop.run_until_complete(event.wait())

    assert TestService.DATA
    assert TestService.DATA == [b"hello server\n"]


async def test_robust_tcp_client(event_loop, aiomisc_socket_factory, localhost):
    condition = asyncio.Condition()

    class TestService(TCPServer):
        DATA = []
        condition: asyncio.Condition

        async def handle_client(
            self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter
        ):
            self.DATA.append(await reader.readline())
            async with self.condition:
                self.condition.notify_all()
            writer.write_eof()
            writer.close()

    class TestClient(RobustTCPClient):
        async def handle_connection(
            self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter
        ) -> None:
            writer.write(b"hello server\n")
            await writer.drain()
            await reader.readexactly(1)

    port, sock = aiomisc_socket_factory()
    services = [
        TestService(sock=sock, condition=condition),
        TestClient(address=localhost, port=port, reconnect_timeout=0.1),
    ]

    async def go():
        async with condition:
            await condition.wait_for(lambda: len(TestService.DATA) >= 3)

    async with aiomisc.entrypoint(*services):
        await asyncio.wait_for(go(), timeout=10)

    assert TestService.DATA
    assert TestService.DATA == [b"hello server\n"] * 3


@pytest.mark.parametrize("client_cert_required", [False, True])
def test_tls_server(client_cert_required, certs, ssl_client_context, localhost):
    class TestService(TLSServer):
        DATA = []
        event: asyncio.Event

        async def start(self):
            await super().start()
            self.event = asyncio.Event()

        async def handle_client(
            self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter
        ):
            print("handle_client")
            self.DATA.append(await reader.readline())
            writer.close()
            self.event.set()

    service = TestService(
        address="127.0.0.1",
        port=0,
        ca=certs / "ca.pem",
        key=certs / "server.key",
        cert=certs / "server.pem",
        require_client_cert=client_cert_required,
    )

    @aiomisc.threaded
    def writer(port):
        with ExitStack() as stack:
            sock = stack.enter_context(
                socket.socket(socket.AF_INET, socket.SOCK_STREAM, 0)
            )

            ssock = stack.enter_context(
                ssl_client_context.wrap_socket(
                    sock, server_hostname="localhost"
                )
            )

            ssock.connect(("127.0.0.1", port))
            count_bytes = ssock.send(b"hello server\n")
            assert count_bytes == len(b"hello server\n")

    @timeout(999)
    async def go():
        await asyncio.wait_for(writer(service.port), timeout=10)
        await service.event.wait()

    with aiomisc.entrypoint(service) as loop:
        assert service.address == localhost
        loop.run_until_complete(go())

    assert TestService.DATA
    assert TestService.DATA == [b"hello server\n"]


async def test_tls_client(event_loop, certs, localhost, aiomisc_socket_factory):
    class TestService(TLSServer):
        DATA = []

        async def handle_client(
            self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter
        ):
            self.DATA.append(await reader.readline())
            writer.close()

    class TestClient(TLSClient):
        event: asyncio.Event

        async def handle_connection(
            self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter
        ) -> None:
            writer.write(b"hello server\n")
            await writer.drain()
            self.event.set()

    event = asyncio.Event()
    port, sock = aiomisc_socket_factory()
    services = [
        TestService(
            sock=sock,
            ca=certs / "ca.pem",
            key=certs / "server.key",
            cert=certs / "server.pem",
        ),
        TestClient(
            address=localhost,
            port=port,
            ca=certs / "ca.pem",
            key=certs / "client.key",
            cert=certs / "client.pem",
            event=event,
        ),
    ]

    async def go():
        await event.wait()

    async with aiomisc.entrypoint(*services):
        await asyncio.wait_for(go(), timeout=10)

    assert TestService.DATA
    assert TestService.DATA == [b"hello server\n"]


async def test_robust_tls_client(
    event_loop, aiomisc_socket_factory, localhost, certs
):
    condition = asyncio.Condition()

    class TestService(TLSServer):
        DATA = []
        condition: asyncio.Condition

        async def handle_client(
            self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter
        ):
            self.DATA.append(await reader.readline())
            async with self.condition:
                self.condition.notify_all()
            writer.close()

    class TestClient(RobustTLSClient):
        async def handle_connection(
            self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter
        ) -> None:
            writer.write(b"hello server\n")
            await writer.drain()
            await reader.readexactly(1)

    port, sock = aiomisc_socket_factory()
    services = [
        TestService(
            sock=sock,
            ca=certs / "ca.pem",
            key=certs / "server.key",
            cert=certs / "server.pem",
            condition=condition,
        ),
        TestClient(
            address=localhost,
            port=port,
            ca=certs / "ca.pem",
            key=certs / "client.key",
            cert=certs / "client.pem",
            reconnect_timeout=0.1,
        ),
    ]

    async def go():
        async with condition:
            await condition.wait_for(lambda: len(TestService.DATA) >= 3)

    async with aiomisc.entrypoint(*services):
        await asyncio.wait_for(go(), timeout=10)

    assert TestService.DATA
    assert TestService.DATA == [b"hello server\n"] * 3


def test_udp_server(aiomisc_socket_factory):
    port, sock = aiomisc_socket_factory(socket.AF_INET, socket.SOCK_DGRAM)

    event = asyncio.Event()

    class TestService(UDPServer):
        DATA = []

        async def handle_datagram(self, data: bytes, addr: tuple) -> None:
            self.DATA.append(data)
            event.set()

    service = TestService("127.0.0.1", sock=sock)

    @aiomisc.threaded
    def writer():
        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)

        with sock:
            sock.sendto(b"hello server\n", ("127.0.0.1", port))

    with aiomisc.entrypoint(service) as loop:
        loop.run_until_complete(asyncio.wait_for(writer(), timeout=10))
        loop.run_until_complete(event.wait())

    assert TestService.DATA
    assert TestService.DATA == [b"hello server\n"]


def test_udp_without_port_or_socket():
    class TestService(UDPServer):
        async def handle_datagram(self, data: bytes, addr: tuple) -> None:
            pass

    with pytest.raises(RuntimeError):
        TestService()


def test_tcp_without_port_or_socket():
    class TestService(TCPServer):
        async def handle_client(
            self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter
        ) -> Any:
            pass

    with pytest.raises(RuntimeError):
        TestService()


@unix_only
def test_udp_socket_server(unix_socket_udp):
    class TestService(UDPServer):
        DATA = []

        async def handle_datagram(self, data: bytes, addr: tuple):
            self.DATA.append(data)

    service = TestService(sock=unix_socket_udp)

    @aiomisc.threaded
    def writer():
        sock = socket.socket(socket.AF_UNIX, socket.SOCK_DGRAM)

        with sock:
            sock.sendto(b"hello server\n", unix_socket_udp.getsockname())

    with aiomisc.entrypoint(service) as loop:
        loop.run_until_complete(asyncio.wait_for(writer(), timeout=10))

    assert TestService.DATA
    assert TestService.DATA == [b"hello server\n"]


@unix_only
def test_tcp_server_unix(unix_socket_tcp):
    class TestService(TCPServer):
        DATA = []

        async def handle_client(
            self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter
        ):
            self.DATA.append(await reader.readline())
            writer.close()

    service = TestService(sock=unix_socket_tcp)

    @aiomisc.threaded
    def writer():
        with socket.socket(socket.AF_UNIX, socket.SOCK_STREAM) as sock:
            sock.connect(unix_socket_tcp.getsockname())
            sock.send(b"hello server\n")

    with aiomisc.entrypoint(service) as loop:
        loop.run_until_complete(asyncio.wait_for(writer(), timeout=10))

    assert TestService.DATA
    assert TestService.DATA == [b"hello server\n"]


def test_aiohttp_service_create_app():
    with pytest.raises(TypeError):

        class NoAsyncCreateApplication(AIOHTTPService):
            def create_application(self):
                return None

    class AsyncCreateApplication(AIOHTTPService):
        async def create_application(self):
            return aiohttp.web.Application()


class AIOHTTPTestApp(AIOHTTPService):
    async def create_application(self):
        return aiohttp.web.Application()


def test_aiohttp_service(aiomisc_unused_port):
    async def http_client():
        session = aiohttp.ClientSession()
        url = f"http://localhost:{aiomisc_unused_port}"

        async with session:
            async with session.get(url) as response:
                return response.status

    service = AIOHTTPTestApp(address="127.0.0.1", port=aiomisc_unused_port)

    with aiomisc.entrypoint(service) as loop:
        response = loop.run_until_complete(
            asyncio.wait_for(http_client(), timeout=10)
        )

    assert response == 404


@unix_only
def test_aiohttp_service_sock(unix_socket_tcp):
    async def http_client():
        conn = aiohttp.UnixConnector(path=unix_socket_tcp.getsockname())
        session = aiohttp.ClientSession(connector=conn)

        async with session:
            async with session.get("http://unix/") as response:
                return response.status

    service = AIOHTTPTestApp(sock=unix_socket_tcp)

    with aiomisc.entrypoint(service) as loop:
        response = loop.run_until_complete(
            asyncio.wait_for(http_client(), timeout=10000)
        )

    assert response == 404


def test_asgi_service_create_app():
    with pytest.raises(TypeError):

        class NoAsyncCreateASGIApp(ASGIHTTPService):
            def create_asgi_app(self) -> ASGIApplicationType:  # type: ignore
                return lambda: None  # type: ignore

    class AsyncCreateASGIApp(ASGIHTTPService):
        async def create_asgi_app(self) -> ASGIApplicationType:
            return fastapi.FastAPI()


class ASGIHTTPTestApp(ASGIHTTPService):
    async def create_asgi_app(self) -> ASGIApplicationType:
        app = fastapi.FastAPI()

        @app.get("/")
        async def root():
            return {"message": "Hello World"}

        return app


def test_aiohttp_service_without_port_or_sock(aiomisc_unused_port):
    with pytest.raises(RuntimeError):
        ASGIHTTPTestApp()


def test_asgi_service(aiomisc_unused_port):
    async def http_client():
        session = aiohttp.ClientSession()
        url = f"http://localhost:{aiomisc_unused_port}"

        async with session:
            async with session.get(url) as response:
                return response.status, await response.json()

    service = ASGIHTTPTestApp(address="127.0.0.1", port=aiomisc_unused_port)

    with aiomisc.entrypoint(service) as loop:
        response, body = loop.run_until_complete(
            asyncio.wait_for(http_client(), timeout=10)
        )

    assert body == {"message": "Hello World"}
    assert response == 200


@unix_only
def test_asgi_service_sock(unix_socket_tcp):
    async def http_client():
        conn = aiohttp.UnixConnector(path=unix_socket_tcp.getsockname())
        session = aiohttp.ClientSession(connector=conn)

        async with session:
            async with session.get("http://unix/") as response:
                return response.status, await response.json()

    service = ASGIHTTPTestApp(sock=unix_socket_tcp)

    with aiomisc.entrypoint(service) as loop:
        response, body = loop.run_until_complete(
            asyncio.wait_for(http_client(), timeout=10000)
        )

    assert body == {"message": "Hello World"}
    assert response == 200


class UvicornTestService(UvicornService):
    async def create_application(self) -> UvicornApplication:
        app = fastapi.FastAPI()

        @app.get("/")
        async def root():
            return {"message": "Hello World"}

        return app


def test_uvicorn_service(aiomisc_unused_port):
    async def http_client():
        session = aiohttp.ClientSession()
        url = f"http://localhost:{aiomisc_unused_port}"

        async with session:
            async with session.get(url) as response:
                return response.status, await response.json()

    service = UvicornTestService(host="127.0.0.1", port=aiomisc_unused_port)

    with aiomisc.entrypoint(service) as loop:
        response, body = loop.run_until_complete(
            asyncio.wait_for(http_client(), timeout=10)
        )

    assert body == {"message": "Hello World"}
    assert response == 200


@unix_only
def test_uvicorn_service_sock(unix_socket_tcp):
    async def http_client():
        conn = aiohttp.UnixConnector(path=unix_socket_tcp.getsockname())
        session = aiohttp.ClientSession(connector=conn)

        async with session:
            async with session.get("http://unix/") as response:
                return response.status, await response.json()

    service = UvicornTestService(sock=unix_socket_tcp)

    with aiomisc.entrypoint(service) as loop:
        response, body = loop.run_until_complete(
            asyncio.wait_for(http_client(), timeout=10000)
        )

    assert body == {"message": "Hello World"}
    assert response == 200


def test_service_events():
    class Initialization(aiomisc.Service):
        async def start(self):
            context = aiomisc.get_context()
            await asyncio.sleep(0.1)
            context["test"] = True

    class Awaiter(aiomisc.Service):
        result = None

        async def start(self):
            Awaiter.result = await self.context["test"]

    services = (Awaiter(), Initialization())

    with aiomisc.entrypoint(*services):
        pass

    assert Awaiter.result


def test_service_events_2():
    class Initialization(aiomisc.Service):
        async def start(self):
            self.context["test"] = True

    class Awaiter(aiomisc.Service):
        result = None

        async def start(self):
            context = aiomisc.get_context()

            await asyncio.sleep(0.1)

            await context["test"]
            Awaiter.result = await context["test"]

    services = (Initialization(), Awaiter())

    with aiomisc.entrypoint(*services):
        pass

    assert Awaiter.result


def test_service_start_event():
    class Sleeper(aiomisc.Service):
        result = False

        async def start(self):
            self.start_event.set()

            await asyncio.sleep(86400)
            Sleeper.result = True

    with aiomisc.entrypoint(Sleeper()):
        pass

    assert not Sleeper.result


def test_service_no_start_event():
    class Sleeper(aiomisc.Service):
        result = False

        async def start(self):
            await asyncio.sleep(1)
            Sleeper.result = True

    with aiomisc.entrypoint(Sleeper()):
        pass

    assert Sleeper.result


def test_context_multiple_set():
    results = []

    async def test():
        context = aiomisc.get_context()

        context["foo"] = True
        await asyncio.sleep(0.1)
        results.append(await context["foo"])

        context["foo"] = False
        await asyncio.sleep(0.1)
        results.append(await context["foo"])

        context["foo"] = None
        await asyncio.sleep(0.1)
        results.append(await context["foo"])

    with aiomisc.entrypoint() as loop:
        loop.run_until_complete(asyncio.wait_for(test(), timeout=10))

    assert results == [True, False, None]


async def test_entrypoint_with_with_async():
    class MyService(aiomisc.service.Service):
        ctx = 0

        async def start(self):
            self.__class__.ctx = 1

        async def stop(self, exc: Exception | None = None) -> None:
            self.__class__.ctx = 2

    service = MyService()
    assert service.ctx == 0

    async with aiomisc.entrypoint(service) as ep:
        assert service.ctx == 1

        with pytest.raises(asyncio.TimeoutError):
            await asyncio.wait_for(ep.closing(), timeout=0.3)

        assert service.ctx == 1

    assert service.ctx == 2


async def test_entrypoint_graceful_shutdown_loop_owner(event_loop):
    event = Event()
    task: Task

    async def func():
        nonlocal event
        await event.wait()

    async def pre_start(**_):
        nonlocal task
        task = get_event_loop().create_task(func())

    async def post_stop(**_):
        nonlocal event, task
        event.set()
        with suppress(asyncio.TimeoutError):
            await asyncio.wait_for(task, timeout=1.0)

    Entrypoint.PRE_START.connect(pre_start)
    Entrypoint.POST_STOP.connect(post_stop)

    async with Entrypoint() as entrypoint:
        assert entrypoint.loop is event_loop

    Entrypoint.PRE_START.disconnect(pre_start)
    Entrypoint.POST_STOP.disconnect(post_stop)

    assert task.done()
    assert not task.cancelled()


class CucumberService(aiomisc.Service):
    async def start(self) -> None:
        return


async def test_service_pickle():
    cucumbers = [CucumberService(index=i) for i in range(100)]
    pickled_cucumbers = pickle.dumps(cucumbers)

    cucumbers.clear()

    for idx, cucumber in enumerate(pickle.loads(pickled_cucumbers)):
        assert cucumber.index == idx


class StorageService(aiomisc.Service):
    INSTANCES: set["StorageService"] = set()

    async def start(self) -> None:
        self.INSTANCES.add(self)

    async def stop(self, exc: Exception | None = None) -> None:
        self.INSTANCES.remove(self)


async def test_add_remove_service(entrypoint: aiomisc.Entrypoint):
    StorageService.INSTANCES.clear()

    service = StorageService()

    await entrypoint.start_services(service)
    assert len(StorageService.INSTANCES) == 1
    assert service in StorageService.INSTANCES

    await entrypoint.stop_services(service)
    assert service not in StorageService.INSTANCES

    assert len(StorageService.INSTANCES) == 0

    for service in map(lambda _: StorageService(), range(10)):
        await entrypoint.start_services(service)

    assert len(StorageService.INSTANCES) == 10


@pytest.mark.parametrize(
    "entrypoint_logging_kwargs,basic_config_kwargs",
    [
        (
            {
                "log_level": LogLevel.info.name,
                "log_format": LogFormat.plain,
                "log_date_format": None,
            },
            {
                "level": LogLevel.info.name,
                "log_format": LogFormat.plain,
                "date_format": None,
            },
        ),
        (
            {
                "log_level": LogLevel.debug,
                "log_format": LogFormat.stream,
                "log_date_format": DateFormat.stream,
            },
            {
                "level": LogLevel.debug,
                "log_format": LogFormat.stream,
                "date_format": DateFormat.stream,
            },
        ),
    ],
)
def test_entrypoint_log_params(entrypoint_logging_kwargs, basic_config_kwargs):
    with mock.patch("aiomisc_log.basic_config") as basic_config_mock:
        with entrypoint(**entrypoint_logging_kwargs):
            pass

        # unbuffered logging is configured on init, buffered on start,
        # unbuffered on stop.
        assert basic_config_mock.call_count == 3
        for call_args, call_kwargs in basic_config_mock.call_args_list:
            for key, value in basic_config_kwargs.items():
                assert call_kwargs[key] == value


@pytest.fixture(scope="session")
def grpc_hello() -> tuple[ModuleType, ModuleType]:
    return grpc.protos_and_services("tests/hello.proto")


def test_grpc_service(localhost, grpc_hello):
    protos, services = grpc_hello
    token = uuid.uuid4().hex

    class Greeter(services.GreeterServicer):  # type: ignore
        async def SayHello(self, request, context):
            return protos.HelloReply(message=f"{request.name}")

    grpc_service = GRPCService(compression=grpc.Compression.Gzip)
    services.add_GreeterServicer_to_server(Greeter(), grpc_service)

    port_future = grpc_service.add_insecure_port(f"{localhost}:0")

    async def go():
        port: int = await port_future

        async with grpc.aio.insecure_channel(f"{localhost}:{port}") as channel:
            stub = services.GreeterStub(channel)
            response = await stub.SayHello(protos.HelloRequest(name=token))

        return response.message

    with aiomisc.entrypoint(grpc_service) as loop:
        result = loop.run_until_complete(go())

    assert result == token
